"""Centroid-based models for supervised classification."""
from __future__ import annotations

from typing import Callable, Generic, Optional, TypeVar

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_is_fitted as sklearn_check_is_fitted

from ..._utils import _classifier_get_classes
from ...exploratory.depth import Depth, ModifiedBandDepth
from ...exploratory.stats import mean, trim_mean
from ...misc.metrics import Metric, PairwiseMetric, l2_distance
from ...representation import FData
from ...representation._typing import NDArrayInt

T = TypeVar("T", bound=FData)


class NearestCentroid(
    BaseEstimator,  # type: ignore
    ClassifierMixin,  # type: ignore
    Generic[T],
):
    """Nearest centroid classifier for functional data.

    Each class is represented by its centroid, with test samples classified to
    the class with the nearest centroid.

    Parameters:
        metric:
            The metric to use when calculating distance between test samples
            and centroids. See the documentation of the metrics module
            for a list of available metrics. L2 distance is used by default.
        centroid:
            The centroids for the samples corresponding to each class is the
            point from which the sum of the distances (according to the metric)
            of all samples that belong to that particular class are minimized.
            By default it is used the usual mean, which minimizes the sum of L2
            distances. This parameter allows change the centroid constructor.
            The function must accept a :class:`FData` with the samples of one
            class and return a :class:`FData` object with only one sample
            representing the centroid.
    Examples:
        Firstly, we will create a toy dataset with 2 classes

        >>> from skfda.datasets import make_sinusoidal_process
        >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
        >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
        ...                               phase_std=.25, random_state=0)
        >>> fd = fd1.concatenate(fd2)
        >>> y = 15*[0] + 15*[1]

        We will fit a Nearest centroids classifier

        >>> from skfda.ml.classification import NearestCentroid
        >>> neigh = NearestCentroid()
        >>> neigh.fit(fd, y)
        NearestCentroid(...)

        We can predict the class of new samples

        >>> neigh.predict(fd[::2]) # Predict labels for even samples
        array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1])

    See also:
        :class:`~skfda.ml.classification.DTMClassifier`
    """

    def __init__(
        self,
        metric: Metric[T] = l2_distance,
        centroid: Callable[[T], T] = mean,
    ):
        self.metric = metric
        self.centroid = centroid

    def fit(self, X: T, y: NDArrayInt) -> NearestCentroid[T]:
        """Fit the model using X as training data and y as target values.

        Args:
            X: FDataGrid with the training data or array matrix with shape
                (n_samples, n_samples) if metric='precomputed'.
            y: Target values of
                shape = (n_samples) or (n_samples, n_outputs).

        Returns:
            self
        """
        classes, y_ind = _classifier_get_classes(y)

        self._classes = classes
        self.centroids_ = self.centroid(X[y_ind == 0])

        for cur_class in range(1, self._classes.size):
            centroid = self.centroid(X[y_ind == cur_class])
            self.centroids_ = self.centroids_.concatenate(centroid)

        return self

    def predict(self, X: T) -> NDArrayInt:
        """Predict the class labels for the provided data.

        Args:
            X: FDataGrid with the test samples.

        Returns:
            Array of shape (n_samples) or
                (n_samples, n_outputs) with class labels for each data sample.
        """
        sklearn_check_is_fitted(self)

        return self._classes[PairwiseMetric(self.metric)(
            X,
            self.centroids_,
        ).argmin(axis=1)
        ]


class DTMClassifier(NearestCentroid[T]):
    """Distance to trimmed means (DTM) classification.

    Test samples are classified to the class that minimizes the distance of
    the observation to the trimmed mean of the group
    :footcite:`fraiman+muniz_2001_trimmed`.

    Parameters:
        proportiontocut:
            Indicates the percentage of functions to remove.
            It is not easy to determine as it varies from dataset to
            dataset.
        depth_method:
            The depth class used to order the data. See the documentation of
            the depths module for a list of available depths. By default it
            is ModifiedBandDepth.
        metric:
            Distance function between two functional objects. See the
            documentation of the metrics module for a list of available
            metrics. L2 distance is used by default.

    Examples:
        Firstly, we will import and split the Berkeley Growth Study dataset

        >>> from skfda.datasets import fetch_growth
        >>> from sklearn.model_selection import train_test_split
        >>> dataset = fetch_growth()
        >>> fd = dataset['data']
        >>> y = dataset['target']
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     fd, y, test_size=0.25, stratify=y, random_state=0)

        We will fit a Distance to trimmed means classifier

        >>> from skfda.ml.classification import DTMClassifier
        >>> clf = DTMClassifier(proportiontocut=0.25)
        >>> clf.fit(X_train, y_train)
        DTMClassifier(...)

        We can predict the class of new samples

        >>> clf.predict(X_test) # Predict labels for test samples
        array([1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
                1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])

        Finally, we calculate the mean accuracy for the test data

        >>> clf.score(X_test, y_test)
        0.875

    See also:
        :class:`~skfda.ml.classification.NearestCentroid`

    References:
        .. footbibliography::

    """

    def __init__(
        self,
        proportiontocut: float,
        depth_method: Optional[Depth[T]] = None,
        metric: Metric[T] = l2_distance,
    ) -> None:
        self.proportiontocut = proportiontocut
        self.depth_method = depth_method

        super().__init__(
            metric,
            centroid=self._centroid,
        )

    def _centroid(self, fdatagrid: T) -> T:
        if self.depth_method is None:
            self.depth_method = ModifiedBandDepth()

        return trim_mean(
            fdatagrid,
            self.proportiontocut,
            depth_method=self.depth_method,
        )



import unittest

import numpy as np

from skfda import FDataGrid
from skfda.exploratory.depth.multivariate import SimplicialDepth
from skfda.exploratory.outliers import (
    MSPlotOutlierDetector,
    directional_outlyingness_stats,
)


class TestsDirectionalOutlyingness(unittest.TestCase):

    def test_directional_outlyingness(self):
        data_matrix = [[[0.3], [0.4], [0.5], [0.6]],
                       [[0.5], [0.6], [0.7], [0.7]],
                       [[0.2], [0.3], [0.4], [0.5]]]
        grid_points = [2, 4, 6, 8]
        fd = FDataGrid(data_matrix, grid_points)
        stats = directional_outlyingness_stats(
            fd, multivariate_depth=SimplicialDepth())
        np.testing.assert_allclose(stats.directional_outlyingness,
                                   np.array([[[0.],
                                              [0.],
                                              [0.],
                                              [0.]],

                                             [[0.5],
                                              [0.5],
                                              [0.5],
                                              [0.5]],

                                             [[-0.5],
                                              [-0.5],
                                              [-0.5],
                                              [-0.5]]]),
                                   rtol=1e-06)
        np.testing.assert_allclose(stats.mean_directional_outlyingness,
                                   np.array([[0.],
                                             [0.5],
                                             [-0.5]]),
                                   rtol=1e-06)
        np.testing.assert_allclose(stats.variation_directional_outlyingness,
                                   np.array([0., 0., 0.]), atol=1e-6)

    def test_asymptotic_formula(self):
        data_matrix = [[1, 1, 2, 3, 2.5, 2],
                       [0.5, 0.5, 1, 2, 1.5, 1],
                       [-1, -1, -0.5, 1, 1, 0.5],
                       [-0.5, -0.5, -0.5, -1, -1, -1]]
        grid_points = [0, 2, 4, 6, 8, 10]
        fd = FDataGrid(data_matrix, grid_points)
        out_detector = MSPlotOutlierDetector(
            _force_asymptotic=True)
        prediction = out_detector.fit_predict(fd)
        np.testing.assert_allclose(prediction,
                                   np.array([1, 1, 1, 1]))


if __name__ == '__main__':
    print()
    unittest.main()



"""Clustering Plots Module."""

from __future__ import annotations

from typing import Optional, Sequence, Tuple, Union

import matplotlib
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.collections import PatchCollection
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from matplotlib.ticker import MaxNLocator
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted
from typing_extensions import Protocol

from ..._utils import _check_compatible_fdata
from ...representation import FData, FDataGrid
from ...representation._typing import NDArrayFloat, NDArrayInt
from ._baseplot import BasePlot
from ._utils import ColorLike, _darken, _set_labels


class ClusteringEstimator(Protocol):

    @property
    def n_clusters(self) -> int:
        pass

    @property
    def cluster_centers_(self) -> FDataGrid:
        pass

    @property
    def labels_(self) -> NDArrayInt:
        pass

    def fit(self, X: FDataGrid) -> ClusteringEstimator:
        pass

    def predict(self, X: FDataGrid) -> NDArrayInt:
        pass


class FuzzyClusteringEstimator(ClusteringEstimator, Protocol):

    def predict_proba(self, X: FDataGrid) -> NDArrayFloat:
        pass


def _plot_clustering_checks(
    estimator: ClusteringEstimator,
    fdata: FData,
    sample_colors: Optional[Sequence[ColorLike]],
    sample_labels: Optional[Sequence[str]],
    cluster_colors: Optional[Sequence[ColorLike]],
    cluster_labels: Optional[Sequence[str]],
    center_colors: Optional[Sequence[ColorLike]],
    center_labels: Optional[Sequence[str]],
) -> None:
    """Check the arguments."""
    if (
        sample_colors is not None
        and len(sample_colors) != fdata.n_samples
    ):
        raise ValueError(
            "sample_colors must contain a color for each sample.",
        )

    if (
        sample_labels is not None
        and len(sample_labels) != fdata.n_samples
    ):
        raise ValueError(
            "sample_labels must contain a label for each sample.",
        )

    if (
        cluster_colors is not None
        and len(cluster_colors) != estimator.n_clusters
    ):
        raise ValueError(
            "cluster_colors must contain a color for each cluster.",
        )

    if (
        cluster_labels is not None
        and len(cluster_labels) != estimator.n_clusters
    ):
        raise ValueError(
            "cluster_labels must contain a label for each cluster.",
        )

    if (
        center_colors is not None
        and len(center_colors) != estimator.n_clusters
    ):
        raise ValueError(
            "center_colors must contain a color for each center.",
        )

    if (
        center_labels is not None
        and len(center_labels) != estimator.n_clusters
    ):
        raise ValueError(
            "centers_labels must contain a label for each center.",
        )


def _get_labels(
    x_label: Optional[str],
    y_label: Optional[str],
    title: Optional[str],
    xlabel_str: str,
) -> Tuple[str, str, str]:
    """
    Get the axes labels.

    Set the arguments *xlabel*, *ylabel*, *title* passed to the plot
    functions :func:`plot_cluster_lines
    <skfda.exploratory.visualization.clustering_plots.plot_cluster_lines>` and
    :func:`plot_cluster_bars
    <skfda.exploratory.visualization.clustering_plots.plot_cluster_bars>`,
    in case they are not set yet.

    Args:
        x_label: Label for the x-axes.
        y_label: Label for the y-axes.
        title: Title for the figure where the clustering results are
            ploted.
        xlabel_str: In case xlabel is None, string to use for the labels
            in the x-axes.

    Returns:
        xlabel: Labels for the x-axes.
        ylabel: Labels for the y-axes.
        title: Title for the figure where the clustering results are
            plotted.
    """
    if x_label is None:
        x_label = xlabel_str

    if y_label is None:
        y_label = "Degree of membership"

    if title is None:
        title = "Degrees of membership of the samples to each cluster"

    return x_label, y_label, title


class ClusterPlot(BasePlot):
    """
    ClusterPlot class.

    Args:
        estimator: estimator used to calculate the
            clusters.
        X: contains the samples which are grouped
            into different clusters.
        fig: figure over which the graphs are plotted in
            case ax is not specified. If None and ax is also None, the figure
            is initialized.
        axes: axis over where the graphs are plotted.
            If None, see param fig.
        n_rows: designates the number of rows of the figure to plot the
            different dimensions of the image. Only specified if fig and
            ax are None.
        n_cols: designates the number of columns of the figure to plot
            the different dimensions of the image. Only specified if fig
            and ax are None.
        sample_labels: contains in order the labels of each
        sample of the fdatagrid.
        cluster_colors: contains in order the colors of each
            cluster the samples of the fdatagrid are classified into.
        cluster_labels: contains in order the names of each
            cluster the samples of the fdatagrid are classified into.
        center_colors: contains in order the colors of each
            centroid of the clusters the samples of the fdatagrid are
            classified into.
        center_labels: contains in order the labels of each
            centroid of the clusters the samples of the fdatagrid are
            classified into.
        center_width: width of the centroid curves.
        colormap: colormap from which the colors of the plot are
            taken. Defaults to `rainbow`.
    """

    def __init__(
        self,
        estimator: ClusteringEstimator,
        fdata: FDataGrid,
        chart: Union[Figure, Axes, None] = None,
        fig: Optional[Figure] = None,
        axes: Union[Axes, Sequence[Axes], None] = None,
        n_rows: Optional[int] = None,
        n_cols: Optional[int] = None,
        sample_labels: Optional[Sequence[str]] = None,
        cluster_colors: Optional[Sequence[ColorLike]] = None,
        cluster_labels: Optional[Sequence[str]] = None,
        center_colors: Optional[Sequence[ColorLike]] = None,
        center_labels: Optional[Sequence[str]] = None,
        center_width: int = 3,
        colormap: matplotlib.colors.Colormap = None,
    ) -> None:

        if colormap is None:
            colormap = plt.cm.get_cmap('rainbow')

        super().__init__(
            chart,
            fig=fig,
            axes=axes,
            n_rows=n_rows,
            n_cols=n_cols,
        )
        self.fdata = fdata
        self.estimator = estimator
        self.sample_labels = sample_labels
        self.cluster_colors = cluster_colors
        self.cluster_labels = cluster_labels
        self.center_colors = center_colors
        self.center_labels = center_labels
        self.center_width = center_width
        self.colormap = colormap

    @property
    def n_subplots(self) -> int:
        return self.fdata.dim_codomain

    @property
    def n_samples(self) -> int:
        return self.fdata.n_samples

    def _plot_clusters(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:
        """Implement the plot of the FDataGrid samples by clusters."""
        _plot_clustering_checks(
            estimator=self.estimator,
            fdata=self.fdata,
            sample_colors=None,
            sample_labels=self.sample_labels,
            cluster_colors=self.cluster_colors,
            cluster_labels=self.cluster_labels,
            center_colors=self.center_colors,
            center_labels=self.center_labels,
        )

        if self.sample_labels is None:
            self.sample_labels = [
                f'$SAMPLE: {i}$' for i in range(self.fdata.n_samples)
            ]

        if self.cluster_colors is None:
            self.cluster_colors = self.colormap(
                np.arange(self.estimator.n_clusters)
                / (self.estimator.n_clusters - 1),
            )

        if self.cluster_labels is None:
            self.cluster_labels = [
                f'$CLUSTER: {i}$' for i in range(self.estimator.n_clusters)
            ]

        if self.center_colors is None:
            self.center_colors = [_darken(c, 0.5) for c in self.cluster_colors]

        if self.center_labels is None:
            self.center_labels = [
                f'$CENTER: {i}$' for i in range(self.estimator.n_clusters)
            ]

        colors_by_cluster = self.cluster_colors[self.labels]

        patches = [
            mpatches.Patch(
                color=self.cluster_colors[i],
                label=self.cluster_labels[i],
            )
            for i in range(self.estimator.n_clusters)
        ]

        artists = [
            axes[j].plot(
                self.fdata.grid_points[0],
                self.fdata.data_matrix[i, :, j],
                c=colors_by_cluster[i],
                label=self.sample_labels[i],
            )
            for j in range(self.fdata.dim_codomain)
            for i in range(self.fdata.n_samples)
        ]

        self.artists = np.array(artists).reshape(
            (self.n_subplots, self.n_samples),
        ).T

        for j in range(self.fdata.dim_codomain):

            for i in range(self.estimator.n_clusters):
                axes[j].plot(
                    self.fdata.grid_points[0],
                    self.estimator.cluster_centers_.data_matrix[i, :, j],
                    c=self.center_colors[i],
                    label=self.center_labels[i],
                    linewidth=self.center_width,
                )
            axes[j].legend(handles=patches)

        _set_labels(self.fdata, fig, axes)

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:

        try:
            check_is_fitted(self.estimator)
            _check_compatible_fdata(
                self.estimator.cluster_centers_,
                self.fdata,
            )
        except NotFittedError:
            self.estimator.fit(self.fdata)

        self.labels = self.estimator.labels_

        self._plot_clusters(fig=fig, axes=axes)


class ClusterMembershipLinesPlot(BasePlot):
    """
    Class ClusterMembershipLinesPlot.

    Args:
        estimator: estimator used to calculate the
            clusters.
        X: contains the samples which are grouped
            into different clusters.
        fig: figure over which the graph is
            plotted in case ax is not specified. If None and ax is also None,
            the figure is initialized.
        axes: axis over where the graph is  plotted.
            If None, see param fig.
        sample_colors: contains in order the colors
            of each sample of the fdatagrid.
        sample_labels: contains in order the labels
            of each sample  of the fdatagrid.
        cluster_labels: contains in order the names of
            each cluster the samples of the fdatagrid are classified into.
        colormap: colormap from which the colors of the
            plot are taken.
        x_label: Label for the x-axis. Defaults to "Cluster".
        y_label: Label for the y-axis. Defaults to
            "Degree of membership".
        title: Title for the figure where the clustering
            results are ploted.
            Defaults to "Degrees of membership of the samples to each cluster".
    """

    def __init__(
        self,
        estimator: FuzzyClusteringEstimator,
        fdata: FDataGrid,
        *,
        chart: Union[Figure, Axes, None] = None,
        fig: Optional[Figure] = None,
        axes: Union[Axes, Sequence[Axes], None] = None,
        sample_colors: Optional[Sequence[ColorLike]] = None,
        sample_labels: Optional[Sequence[str]] = None,
        cluster_labels: Optional[Sequence[str]] = None,
        colormap: matplotlib.colors.Colormap = None,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        title: Optional[str] = None,
    ) -> None:

        if colormap is None:
            colormap = plt.cm.get_cmap('rainbow')

        super().__init__(
            chart,
            fig=fig,
            axes=axes,
        )
        self.fdata = fdata
        self.estimator = estimator
        self.sample_labels = sample_labels
        self.sample_colors = sample_colors
        self.cluster_labels = cluster_labels
        self.x_label = x_label
        self.y_label = y_label
        self.title = title
        self.colormap = colormap

    @property
    def n_samples(self) -> int:
        return self.fdata.n_samples

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:

        try:
            check_is_fitted(self.estimator)
            _check_compatible_fdata(
                self.estimator.cluster_centers_,
                self.fdata,
            )
        except NotFittedError:
            self.estimator.fit(self.fdata)

        membership = self.estimator.predict_proba(self.fdata)

        _plot_clustering_checks(
            estimator=self.estimator,
            fdata=self.fdata,
            sample_colors=self.sample_colors,
            sample_labels=self.sample_labels,
            cluster_colors=None,
            cluster_labels=self.cluster_labels,
            center_colors=None,
            center_labels=None,
        )

        x_label, y_label, title = _get_labels(
            self.x_label,
            self.y_label,
            self.title,
            "Cluster",
        )

        if self.sample_colors is None:
            self.cluster_colors = self.colormap(
                np.arange(self.estimator.n_clusters)
                / (self.estimator.n_clusters - 1),
            )
            labels_by_cluster = self.estimator.labels_
            self.sample_colors = self.cluster_colors[labels_by_cluster]

        if self.sample_labels is None:
            self.sample_labels = [
                f'$SAMPLE: {i}$'
                for i in range(self.fdata.n_samples)
            ]

        if self.cluster_labels is None:
            self.cluster_labels = [
                f'${i}$'
                for i in range(self.estimator.n_clusters)
            ]

        axes[0].get_xaxis().set_major_locator(MaxNLocator(integer=True))
        self.artists = np.array([
            axes[0].plot(
                np.arange(self.estimator.n_clusters),
                membership[i],
                label=self.sample_labels[i],
                color=self.sample_colors[i],
            )
            for i in range(self.fdata.n_samples)
        ])

        axes[0].set_xticks(np.arange(self.estimator.n_clusters))
        axes[0].set_xticklabels(self.cluster_labels)
        axes[0].set_xlabel(x_label)
        axes[0].set_ylabel(y_label)

        fig.suptitle(title)


class ClusterMembershipPlot(BasePlot):
    """
    Class ClusterMembershipPlot.

    Args:
        estimator: estimator used to calculate the
            clusters.
        X: contains the samples which are grouped
            into different clusters.
        fig: figure over which the graph is
            plotted in case ax is not specified. If None and ax is also None,
            the figure is initialized.
        axes: axis over where the graph is  plotted.
            If None, see param fig.
        sample_colors: contains in order the colors
            of each sample of the fdatagrid.
        sample_labels: contains in order the labels
            of each sample  of the fdatagrid.
        cluster_labels: contains in order the names of
            each cluster the samples of the fdatagrid are classified into.
        colormap: colormap from which the colors of the
            plot are taken.
        x_label: Label for the x-axis. Defaults to "Cluster".
        y_label: Label for the y-axis. Defaults to
            "Degree of membership".
        title: Title for the figure where the clustering
            results are ploted.
            Defaults to "Degrees of membership of the samples to each cluster".
    """

    def __init__(
        self,
        estimator: FuzzyClusteringEstimator,
        fdata: FData,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Union[Axes, Sequence[Axes], None] = None,
        sort: int = -1,
        sample_labels: Optional[Sequence[str]] = None,
        cluster_colors: Optional[Sequence[ColorLike]] = None,
        cluster_labels: Optional[Sequence[str]] = None,
        colormap: matplotlib.colors.Colormap = None,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        title: Optional[str] = None,
    ) -> None:

        if colormap is None:
            colormap = plt.cm.get_cmap('rainbow')

        super().__init__(
            chart,
            fig=fig,
            axes=axes,
        )
        self.fdata = fdata
        self.estimator = estimator
        self.sample_labels = sample_labels
        self.cluster_colors = cluster_colors
        self.cluster_labels = cluster_labels
        self.x_label = x_label
        self.y_label = y_label
        self.title = title
        self.colormap = colormap
        self.sort = sort

    @property
    def n_samples(self) -> int:
        return self.fdata.n_samples

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:

        self.artists = np.full(
            (self.n_samples, self.n_subplots),
            None,
            dtype=Artist,
        )

        try:
            check_is_fitted(self.estimator)
            _check_compatible_fdata(
                self.estimator.cluster_centers_,
                self.fdata,
            )
        except NotFittedError:
            self.estimator.fit(self.fdata)

        membership = self.estimator.predict_proba(self.fdata)

        if self.sort < -1 or self.sort >= self.estimator.n_clusters:
            raise ValueError(
                "The sorting number must belong to "
                "the interval [-1, n_clusters)",
            )

        _plot_clustering_checks(
            estimator=self.estimator,
            fdata=self.fdata,
            sample_colors=None,
            sample_labels=self.sample_labels,
            cluster_colors=self.cluster_colors,
            cluster_labels=self.cluster_labels,
            center_colors=None,
            center_labels=None,
        )

        x_label, y_label, title = _get_labels(
            self.x_label,
            self.y_label,
            self.title,
            "Sample",
        )

        if self.sample_labels is None:
            self.sample_labels = np.arange(self.fdata.n_samples)

        if self.cluster_colors is None:
            self.cluster_colors = self.colormap(
                np.arange(self.estimator.n_clusters)
                / (self.estimator.n_clusters - 1),
            )

        if self.cluster_labels is None:
            self.cluster_labels = [
                f'$CLUSTER: {i}$'
                for i in range(self.estimator.n_clusters)
            ]

        patches = [
            mpatches.Patch(
                color=self.cluster_colors[i],
                label=self.cluster_labels[i],
            )
            for i in range(self.estimator.n_clusters)
        ]

        if self.sort == -1:
            labels_dim = membership
        else:
            sample_indices = np.argsort(-membership[:, self.sort])
            self.sample_labels = np.copy(self.sample_labels[sample_indices])
            labels_dim = np.copy(membership[sample_indices])

            temp_labels = np.copy(labels_dim[:, 0])
            labels_dim[:, 0] = labels_dim[:, self.sort]
            labels_dim[:, self.sort] = temp_labels

            temp_color = np.copy(self.cluster_colors[0])
            self.cluster_colors[0] = self.cluster_colors[self.sort]
            self.cluster_colors[self.sort] = temp_color

        conc = np.zeros((self.fdata.n_samples, 1))
        labels_dim = np.concatenate((conc, labels_dim), axis=-1)
        bars = [
            axes[0].bar(
                np.arange(self.fdata.n_samples),
                labels_dim[:, i + 1],
                bottom=np.sum(labels_dim[:, :(i + 1)], axis=1),
                color=self.cluster_colors[i],
            )
            for i in range(self.estimator.n_clusters)
        ]

        for b in bars:
            b.remove()
            b.figure = None

        for i in range(self.n_samples):
            collection = PatchCollection(
                [
                    Rectangle(
                        bar.patches[i].get_xy(),
                        bar.patches[i].get_width(),
                        bar.patches[i].get_height(),
                        color=bar.patches[i].get_facecolor(),
                    ) for bar in bars
                ],
                match_original=True,
            )
            axes[0].add_collection(collection)
            self.artists[i, 0] = collection

        fig.canvas.draw()

        axes[0].set_xticks(np.arange(self.fdata.n_samples))
        axes[0].set_xticklabels(self.sample_labels)
        axes[0].set_xlabel(x_label)
        axes[0].set_ylabel(y_label)
        axes[0].legend(handles=patches)

        fig.suptitle(title)



"""Operators applicable to functional data."""
from ._identity import Identity
from ._integral_transform import IntegralTransform
from ._linear_differential_operator import LinearDifferentialOperator
from ._operators import (
    MatrixOperator,
    Operator,
    gramian_matrix,
    gramian_matrix_optimization,
)
from ._srvf import SRSF



# -*- coding: utf-8 -*-
"""Kernel smoother functions.

This module includes the most commonly used kernel smoother methods for FDA.
 So far only non parametric methods are implemented because we are only
 relying on a discrete representation of functional data.

"""
import abc

import numpy as np

from ...misc import kernels
from ._linear import _LinearSmoother

__author__ = "Miguel Carbajo Berrocal"
__email__ = "miguel.carbajo@estudiante.uam.es"


class _LinearKernelSmoother(_LinearSmoother):

    def __init__(self, *, smoothing_parameter=None,
                 kernel=kernels.normal, weights=None,
                 output_points=None):
        self.smoothing_parameter = smoothing_parameter
        self.kernel = kernel
        self.weights = weights
        self.output_points = output_points
        self._cv = False  # For testing purposes only

    def _hat_matrix(self, input_points, output_points):
        return self._hat_matrix_function(
            input_points=input_points[0],
            output_points=output_points[0],
            smoothing_parameter=self.smoothing_parameter,
            kernel=self.kernel,
            weights=self.weights,
            _cv=self._cv
        )

    def _hat_matrix_function(self, *, input_points, output_points,
                             smoothing_parameter,
                             kernel, weights, _cv=False):

        # Time deltas
        delta_x = np.abs(np.subtract.outer(output_points, input_points))

        # Obtain the non-normalized matrix
        matrix = self._hat_matrix_function_not_normalized(
            delta_x=delta_x,
            smoothing_parameter=smoothing_parameter,
            kernel=kernel)

        # Adjust weights
        if weights is not None:
            matrix = matrix * weights

        # Set diagonal to cero if requested (for testing purposes only)
        if _cv:
            np.fill_diagonal(matrix, 0)

        # Renormalize weights
        rs = np.sum(matrix, 1)
        rs[rs == 0] = 1
        return (matrix.T / rs).T

    @abc.abstractmethod
    def _hat_matrix_function_not_normalized(self, *, delta_x,
                                            smoothing_parameter, kernel):
        pass

    def _more_tags(self):
        return {
            'X_types': []
        }


class NadarayaWatsonSmoother(_LinearKernelSmoother):
    r"""Nadaraya-Watson smoothing method.

    It is a linear kernel smoothing method.
    Uses an smoothing matrix :math:`\hat{H}` for the discretisation
    points in argvals by the Nadaraya-Watson estimator. The smoothed
    values :math:`\hat{X}` at the points :math:`(t_1', t_2', ..., t_m')`
    can be calculated as :math:`\hat{X} = \hat{H}X` where :math:`X` is the
    vector of observations at the points of discretisation
    :math:`(t_1, t_2, ..., t_n)` and

    .. math::
        \hat{H}_{i,j} = \frac{K\left(\frac{t_j-t_i'}{h}\right)}{\sum_{k=1}^{
        n}K\left(\frac{t_k-t_i'}{h}\right)}

    where :math:`K(\cdot)` is a kernel function and :math:`h` the kernel
    window width or smoothing parameter.

    Args:
        argvals (ndarray): Vector of discretisation points.
        smoothing_parameter (float, optional): Window width of the kernel
            (also called h or bandwidth).
        kernel (function, optional): kernel function. By default a normal
            kernel.
        weights (ndarray, optional): Case weights matrix (in order to modify
            the importance of each point).
        output_points (ndarray, optional): The output points. If ommited,
            the input points are used.

    Examples:
        >>> from skfda import FDataGrid
        >>> fd = FDataGrid(grid_points=[1, 2, 4, 5, 7],
        ...                data_matrix=[[1, 2, 3, 4, 5]])
        >>> smoother = NadarayaWatsonSmoother(smoothing_parameter=3.5)
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 2.42],
                [ 2.61],
                [ 3.03],
                [ 3.24],
                [ 3.65]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.294, 0.282, 0.204, 0.153, 0.068],
               [ 0.249, 0.259, 0.22 , 0.179, 0.093],
               [ 0.165, 0.202, 0.238, 0.229, 0.165],
               [ 0.129, 0.172, 0.239, 0.249, 0.211],
               [ 0.073, 0.115, 0.221, 0.271, 0.319]])
        >>> smoother = NadarayaWatsonSmoother(smoothing_parameter=2)
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.84],
                [ 2.18],
                [ 3.09],
                [ 3.55],
                [ 4.28]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.425, 0.375, 0.138, 0.058, 0.005],
               [ 0.309, 0.35 , 0.212, 0.114, 0.015],
               [ 0.103, 0.193, 0.319, 0.281, 0.103],
               [ 0.046, 0.11 , 0.299, 0.339, 0.206],
               [ 0.006, 0.022, 0.163, 0.305, 0.503]])

        The output points can be changed:

        >>> smoother = NadarayaWatsonSmoother(
        ...                 smoothing_parameter=2,
        ...                 output_points=[1, 2, 3, 4, 5, 6, 7])
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.84],
                [ 2.18],
                [ 2.61],
                [ 3.09],
                [ 3.55],
                [ 3.95],
                [ 4.28]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.425,  0.375,  0.138,  0.058,  0.005],
               [ 0.309,  0.35 ,  0.212,  0.114,  0.015],
               [ 0.195,  0.283,  0.283,  0.195,  0.043],
               [ 0.103,  0.193,  0.319,  0.281,  0.103],
               [ 0.046,  0.11 ,  0.299,  0.339,  0.206],
               [ 0.017,  0.053,  0.238,  0.346,  0.346],
               [ 0.006,  0.022,  0.163,  0.305,  0.503]])

    References:
        Wasserman, L. (2006).   Local Regression.
        In *All of Nonparametric Statistics* (pp. 71). Springer.
    """

    def _hat_matrix_function_not_normalized(self, *, delta_x,
                                            smoothing_parameter,
                                            kernel):
        if smoothing_parameter is None:
            smoothing_parameter = np.percentile(delta_x, 15)

        k = kernel(delta_x / smoothing_parameter)
        return k


class LocalLinearRegressionSmoother(_LinearKernelSmoother):
    r"""Local linear regression smoothing method.

    It is a linear kernel smoothing method.
    Uses an smoothing matrix :math:`\hat{H}` for the discretisation
    points in argvals by the local linear regression estimator. The smoothed
    values :math:`\hat{X}` at the points :math:`(t_1', t_2', ..., t_m')`
    can be calculated as :math:`\hat{X} = \hat{H}X` where :math:`X` is the
    vector of observations at the points of discretisation
    :math:`(t_1, t_2, ..., t_n)` and

    .. math::
        \hat{H}_{i,j} = \frac{b_j(t_i')}{\sum_{k=1}^{n}b_k(t_i')}

    .. math::
        b_j(t') = K\left(\frac{t_j - t'}{h}\right) S_{n,2}(t') -
        (t_j - t')S_{n,1}(t')

    .. math::
        S_{n,k}(t') = \sum_{j=1}^{n}K\left(\frac{t_j-t'}{h}\right)(t_j-t')^k

    where :math:`K(\cdot)` is a kernel function and :math:`h` the kernel
    window width.

    Args:
        argvals (ndarray): Vector of discretisation points.
        smoothing_parameter (float, optional): Window width of the kernel
            (also called h or bandwidth).
        kernel (function, optional): kernel function. By default a normal
            kernel.
        weights (ndarray, optional): Case weights matrix (in order to modify
            the importance of each point).
        output_points (ndarray, optional): The output points. If ommited,
            the input points are used.

    Examples:
        >>> from skfda import FDataGrid
        >>> fd = FDataGrid(grid_points=[1, 2, 4, 5, 7],
        ...                data_matrix=[[1, 2, 3, 4, 5]])
        >>> smoother = LocalLinearRegressionSmoother(smoothing_parameter=3.5)
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.13],
                [ 1.36],
                [ 3.29],
                [ 4.27],
                [ 5.08]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.614,  0.429,  0.077, -0.03 , -0.09 ],
               [ 0.381,  0.595,  0.168, -0.   , -0.143],
               [-0.104,  0.112,  0.697,  0.398, -0.104],
               [-0.147, -0.036,  0.392,  0.639,  0.152],
               [-0.095, -0.079,  0.117,  0.308,  0.75 ]])
        >>> smoother = LocalLinearRegressionSmoother(smoothing_parameter=2)
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.11],
                [ 1.41],
                [ 3.31],
                [ 4.04],
                [ 5.04]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.714,  0.386, -0.037, -0.053, -0.01 ],
               [ 0.352,  0.724,  0.045, -0.081, -0.04 ],
               [-0.078,  0.052,  0.74 ,  0.364, -0.078],
               [-0.07 , -0.067,  0.36 ,  0.716,  0.061],
               [-0.012, -0.032, -0.025,  0.154,  0.915]])

        The output points can be changed:

        >>> smoother = LocalLinearRegressionSmoother(
        ...                 smoothing_parameter=2,
        ...                 output_points=[1, 2, 3, 4, 5, 6, 7])
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.11],
                [ 1.41],
                [ 1.81],
                [ 3.31],
                [ 4.04],
                [ 5.35],
                [ 5.04]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.714,  0.386, -0.037, -0.053, -0.01 ],
               [ 0.352,  0.724,  0.045, -0.081, -0.04 ],
               [-0.084,  0.722,  0.722, -0.084, -0.278],
               [-0.078,  0.052,  0.74 ,  0.364, -0.078],
               [-0.07 , -0.067,  0.36 ,  0.716,  0.061],
               [-0.098, -0.202, -0.003,  0.651,  0.651],
               [-0.012, -0.032, -0.025,  0.154,  0.915]])

    References:
        Wasserman, L. (2006).   Local Regression.
        In *All of Nonparametric Statistics* (pp. 77). Springer.
    """

    def _hat_matrix_function_not_normalized(self, *, delta_x,
                                            smoothing_parameter, kernel):
        k = kernel(delta_x / smoothing_parameter)

        s1 = np.sum(k * delta_x, 1, keepdims=True)  # S_n_1
        s2 = np.sum(k * delta_x ** 2, 1, keepdims=True)  # S_n_2
        b = (k * (s2 - delta_x * s1))  # b_i(x_j)
        return b


class KNeighborsSmoother(_LinearKernelSmoother):
    r"""K-nearest neighbour kernel smoother.

    It is a linear kernel smoothing method.
    Uses an smoothing matrix S for the discretisation points in argvals by
    the :math:`k` nearest neighbours estimator.

    The smoothed values :math:`\hat{X}` at the points
    :math:`(t_1', t_2', ..., t_m')` can be calculated as
    :math:`\hat{X} = \hat{H}X` where :math:`X` is the vector of observations
    at the points of discretisation :math:`(t_1, t_2, ..., t_n)` and

    .. math::

        H_{i,j} =\frac{K\left(\frac{t_j-t_i'}{h_{ik}}\right)}{\sum_{r=1}^n
        K\left(\frac{t_r-t_i'}{h_{ik}}\right)}

    :math:`K(\cdot)` is a kernel function and :math:`h_{ik}` the is the
    distance from :math:`t_i'` to the 𝑘-th nearest neighbor of :math:`t_i'`.

    Usually used with the uniform kernel, it takes the average of the closest k
    points to a given point.

    Args:
        argvals (ndarray): Vector of discretisation points.
        smoothing_parameter (int, optional): Number of nearest neighbours. By
            default it takes the 5% closest points.
        kernel (function, optional): kernel function. By default a uniform
            kernel to perform a 'usual' k nearest neighbours estimation.
        weights (ndarray, optional): Case weights matrix (in order to modify
            the importance of each point).
        output_points (ndarray, optional): The output points. If ommited,
            the input points are used.

    Examples:
        >>> from skfda import FDataGrid
        >>> fd = FDataGrid(grid_points=[1, 2, 4, 5, 7],
        ...                data_matrix=[[1, 2, 3, 4, 5]])
        >>> smoother = KNeighborsSmoother(smoothing_parameter=2)
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.5],
                [ 1.5],
                [ 3.5],
                [ 3.5],
                [ 4.5]]])

        >>> smoother.hat_matrix().round(3)
        array([[ 0.5, 0.5, 0. , 0. , 0. ],
               [ 0.5, 0.5, 0. , 0. , 0. ],
               [ 0. , 0. , 0.5, 0.5, 0. ],
               [ 0. , 0. , 0.5, 0.5, 0. ],
               [ 0. , 0. , 0. , 0.5, 0.5]])

        In case there are two points at the same distance it will take both.

        >>> fd = FDataGrid(grid_points=[1, 2, 3, 5, 7],
        ...                data_matrix=[[1, 2, 3, 4, 5]])
        >>> smoother = KNeighborsSmoother(smoothing_parameter=2)
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.5],
                [ 2. ],
                [ 2.5],
                [ 4. ],
                [ 4.5]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.5  , 0.5  , 0.   , 0.   , 0.   ],
               [ 0.333, 0.333, 0.333, 0.   , 0.   ],
               [ 0.   , 0.5  , 0.5  , 0.   , 0.   ],
               [ 0.   , 0.   , 0.333, 0.333, 0.333],
               [ 0.   , 0.   , 0.   , 0.5  , 0.5  ]])

        The output points can be changed:

        >>> smoother = KNeighborsSmoother(
        ...                 smoothing_parameter=2,
        ...                 output_points=[1, 2, 3, 4, 5, 6, 7])
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.5],
                [ 2. ],
                [ 2.5],
                [ 3.5],
                [ 4. ],
                [ 4.5],
                [ 4.5]]])

        >>> smoother.hat_matrix().round(3)
        array([[ 0.5  ,  0.5  ,  0.   ,  0.   ,  0.   ],
               [ 0.333,  0.333,  0.333,  0.   ,  0.   ],
               [ 0.   ,  0.5  ,  0.5  ,  0.   ,  0.   ],
               [ 0.   ,  0.   ,  0.5  ,  0.5  ,  0.   ],
               [ 0.   ,  0.   ,  0.333,  0.333,  0.333],
               [ 0.   ,  0.   ,  0.   ,  0.5  ,  0.5  ],
               [ 0.   ,  0.   ,  0.   ,  0.5  ,  0.5  ]])

    References:
        Frederic Ferraty, Philippe Vieu (2006).  kNN Estimator.
        In *Nonparametric Functional Data Analysis: Theory and Practice*
        (pp. 116). Springer.
    """

    def __init__(self, *, smoothing_parameter=None,
                 kernel=kernels.uniform, weights=None,
                 output_points=None):
        super().__init__(
            smoothing_parameter=smoothing_parameter,
            kernel=kernel,
            weights=weights,
            output_points=output_points
        )

    def _hat_matrix_function_not_normalized(self, *, delta_x,
                                            smoothing_parameter, kernel):

        input_points_len = delta_x.shape[1]

        if smoothing_parameter is None:
            smoothing_parameter = np.floor(np.percentile(
                range(1, input_points_len), 5))
        elif smoothing_parameter <= 0:
            raise ValueError('h must be greater than 0')

        # Tolerance to avoid points landing outside the kernel window due to
        # computation error
        tol = 1.0e-19

        # For each row in the distances matrix, it calculates the furthest
        # point within the k nearest neighbours
        vec = np.percentile(delta_x, smoothing_parameter
                            / input_points_len * 100,
                            axis=1, interpolation='lower') + tol

        rr = kernel((delta_x.T / vec).T)

        return rr



"""Tests for the metrics module."""

import unittest

import numpy as np

from skfda import FDataBasis, FDataGrid
from skfda.datasets import make_multimodal_samples
from skfda.misc.metrics import (
    l1_norm,
    l2_distance,
    l2_norm,
    linf_norm,
    lp_norm,
)
from skfda.representation.basis import Monomial


class TestLp(unittest.TestCase):
    """Test the lp norms and distances."""

    def setUp(self) -> None:
        """Create a few functional data objects."""
        grid_points = [1, 2, 3, 4, 5]
        self.fd = FDataGrid(
            [
                [2, 3, 4, 5, 6],
                [1, 4, 9, 16, 25],
            ],
            grid_points=grid_points,
        )
        basis = Monomial(n_basis=3, domain_range=(1, 5))
        self.fd_basis = FDataBasis(basis, [[1, 1, 0], [0, 0, 1]])
        self.fd_vector_valued = self.fd.concatenate(
            self.fd,
            as_coordinates=True,
        )
        self.fd_surface = make_multimodal_samples(
            n_samples=3,
            dim_domain=2,
            random_state=0,
        )

    def test_lp_norm_grid(self) -> None:
        """Test that the Lp norms work with FDataGrid."""
        np.testing.assert_allclose(
            l1_norm(self.fd),
            [16.0, 41.33333333],
        )

        np.testing.assert_allclose(
            l2_norm(self.fd),
            [8.326664, 25.006666],
        )

        np.testing.assert_allclose(
            lp_norm(self.fd, p=3),
            [6.839904, 22.401268],
        )

        np.testing.assert_allclose(
            linf_norm(self.fd),
            [6, 25],
        )

    def test_lp_norm_basis(self) -> None:
        """Test that the L2 norm works with FDataBasis."""
        np.testing.assert_allclose(
            l2_norm(self.fd_basis),
            [8.326664, 24.996],
        )

    def test_lp_norm_vector_valued(self) -> None:
        """Test that the Lp norms work with vector-valued FDataGrid."""
        np.testing.assert_allclose(
            l1_norm(self.fd_vector_valued),
            [32.0, 82.666667],
        )
        np.testing.assert_allclose(
            linf_norm(self.fd_vector_valued),
            [6, 25],
        )

    def test_lp_norm_surface_inf(self) -> None:
        """Test that the Linf norm works with multidimensional domains."""
        np.testing.assert_allclose(
            lp_norm(self.fd_surface, p=np.inf).round(5),
            [0.99994, 0.99793, 0.99868],
        )

    def test_lp_norm_surface(self) -> None:
        """Test that integration of surfaces has not been implemented."""
        # Integration of surfaces not implemented, add test case after
        # implementation
        self.assertEqual(lp_norm(self.fd_surface, p=1), NotImplemented)

    def test_lp_error_dimensions(self) -> None:
        """Test error on metric between different kind of objects."""
        # Case internal arrays
        with np.testing.assert_raises(ValueError):
            l2_distance(self.fd, self.fd_surface)

        with np.testing.assert_raises(ValueError):
            l2_distance(self.fd, self.fd_vector_valued)

        with np.testing.assert_raises(ValueError):
            l2_distance(self.fd_surface, self.fd_vector_valued)

    def test_lp_error_domain_ranges(self) -> None:
        """Test error on metric between objects with different domains."""
        grid_points = [2, 3, 4, 5, 6]
        fd2 = FDataGrid(
            [
                [2, 3, 4, 5, 6],
                [1, 4, 9, 16, 25],
            ],
            grid_points=grid_points,
        )

        with np.testing.assert_raises(ValueError):
            l2_distance(self.fd, fd2)

    def test_lp_error_grid_points(self) -> None:
        """Test error on metric for FDataGrids with different grid points."""
        grid_points = [1, 2, 4, 4.3, 5]
        fd2 = FDataGrid(
            [
                [2, 3, 4, 5, 6],
                [1, 4, 9, 16, 25],
            ],
            grid_points=grid_points,
        )

        with np.testing.assert_raises(ValueError):
            l2_distance(self.fd, fd2)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import numbers
from typing import Callable, Optional, Sequence, Tuple, Union, cast

import numpy as np
from numpy import polyder, polyint, polymul, polyval

import scipy.integrate
from scipy.interpolate import PPoly

from ..._utils import _same_domain
from ...representation import FData, FDataGrid
from ...representation._typing import DomainRangeLike
from ...representation.basis import (
    BSpline,
    Constant,
    FDataBasis,
    Fourier,
    Monomial,
)
from ._operators import Operator, gramian_matrix_optimization

Order = int

WeightSequence = Sequence[
    Union[
        float,
        Callable[[np.ndarray], np.ndarray],
    ],
]


class LinearDifferentialOperator(
    Operator[FData, Callable[[np.ndarray], np.ndarray]],
):
    r"""
    Defines the structure of a linear differential operator function system.

    .. math::
        Lx(t) = b_0(t) x(t) + b_1(t) x'(x) +
                \dots + b_{n-1}(t) d^{n-1}(x(t)) + b_n(t) d^n(x(t))

    Can only be applied to functional data, as multivariate data has no
    derivatives.

    You have to provide either order or weights.
    If both are provided, it will raise an error.
    If a positional argument is supplied it will be considered the
    order if it is an integral type and the weights otherwise.

    Parameters:
        order (int, optional): the order of the operator. It's the highest
                derivative order of the operator

        weights (list, optional): A FDataBasis objects list of length
                order + 1 items

        domain_range (tuple or list of tuples, optional): Definition
                     of the interval where the weight functions are
                     defined. If the functional weights are specified
                     and this is not, takes the domain range from them.
                     Otherwise, defaults to (0,1).

    Attributes:
        weights (list):  A list of callables.

    Examples:
        Create a linear differential operator that penalizes the second
        derivative (acceleration)

        >>> from skfda.misc.operators import LinearDifferentialOperator
        >>> from skfda.representation.basis import (FDataBasis,
        ...                                         Monomial, Constant)
        >>>
        >>> LinearDifferentialOperator(2)
        LinearDifferentialOperator(
            weights=(0, 0, 1),
        )

        Create a linear differential operator that penalizes three times
        the second derivative (acceleration) and twice the first (velocity).

        >>> LinearDifferentialOperator(weights=[0, 2, 3])
        LinearDifferentialOperator(
            weights=(0, 2, 3),
        )

        Create a linear differential operator with non-constant weights.

        >>> constant = Constant()
        >>> monomial = Monomial(domain_range=(0, 1), n_basis=3)
        >>> fdlist = [FDataBasis(constant, [0.]),
        ...           FDataBasis(constant, [0.]),
        ...           FDataBasis(monomial, [1., 2., 3.])]
        >>> LinearDifferentialOperator(weights=fdlist)
        LinearDifferentialOperator(
        weights=(FDataBasis(
                basis=Constant(domain_range=((0, 1),), n_basis=1),
                coefficients=[[ 0.]],
                ...),
            FDataBasis(
                basis=Constant(domain_range=((0, 1),), n_basis=1),
                coefficients=[[ 0.]],
                ...),
            FDataBasis(
                basis=Monomial(domain_range=((0, 1),), n_basis=3),
                coefficients=[[ 1. 2. 3.]],
                ...)),
        )

    """

    def __init__(
        self,
        order_or_weights: Union[Order, WeightSequence, None] = None,
        *,
        order: Optional[int] = None,
        weights: Optional[WeightSequence] = None,
        domain_range: Optional[DomainRangeLike] = None,
    ) -> None:

        num_args = sum(
            a is not None for a in (order_or_weights, order, weights)
        )

        if num_args > 1:
            raise ValueError(
                "You have to provide the order or the weights, not both.",
            )

        if order_or_weights is not None:
            if isinstance(order_or_weights, numbers.Integral):
                order = int(order_or_weights)
            else:
                assert isinstance(order_or_weights, Sequence)
                weights = order_or_weights

        if order is None and weights is None:
            weights = (0,)

        elif order is not None:
            if order < 0:
                raise ValueError("Order should be an non-negative integer.")

            weights = tuple(
                0 if (i < order) else 1
                for i in range(order + 1)
            )

        assert weights is not None
        if len(weights) == 0:
            raise ValueError("You have to provide one weight at least.")

        # Check domain ranges
        for w in weights:
            w_domain_range = getattr(w, "domain_range", None)

            if w_domain_range is not None:
                if domain_range is None:
                    domain_range = w_domain_range
                elif not np.array_equal(w_domain_range, domain_range):
                    raise ValueError(
                        "Weights with wrong domain range.",
                    )
        self.weights = tuple(weights)

    def __repr__(self) -> str:

        return (
            f"{self.__class__.__name__}(\n"
            f"\tweights={self.weights},\n"
            f")"
        ).replace('\n', '\n    ')

    def __eq__(self, other: object) -> bool:

        return (
            isinstance(other, LinearDifferentialOperator)
            and self.weights == other.weights
        )

    def constant_weights(self) -> Optional[np.ndarray]:
        """
        Return constant weights.

        Return the scalar weights of the linear differential operator if they
        are constant basis.
        Otherwise, return None.

        This function is mostly useful for basis which want to override
        the _penalty method in order to use an analytical expression
        for constant weights.

        """
        weights = np.array(self.weights)

        return None if weights.dtype == np.object_ else weights

    def __call__(self, f: FData) -> Callable[[np.ndarray], np.ndarray]:
        """Return the function that results of applying the operator."""
        function_derivatives = [
            f.derivative(order=i) for i, _ in enumerate(self.weights)
        ]

        def applied_linear_diff_op(
            t: np.ndarray,
        ) -> np.ndarray:
            return sum(
                (w(t) if callable(w) else w)
                * function_derivatives[i](t)
                for i, w in enumerate(self.weights)
            )

        return applied_linear_diff_op


#############################################################
#
# Optimized implementations of gramian matrix for each basis.
#
#############################################################


@gramian_matrix_optimization.register
def constant_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: Constant,
) -> np.ndarray:
    """Optimized version for Constant basis."""
    coefs = linear_operator.constant_weights()
    if coefs is None:
        return NotImplemented

    return np.array([[
        coefs[0] ** 2
        * (basis.domain_range[0][1] - basis.domain_range[0][0]),
    ]])


def _monomial_evaluate_constant_linear_diff_op(
    basis: Monomial,
    weights: np.ndarray,
) -> np.ndarray:
    """Evaluate constant weights over the monomial basis."""
    max_derivative = len(weights) - 1

    seq = np.arange(basis.n_basis)
    coef_mat = np.linspace(
        seq,
        seq - max_derivative + 1,
        max_derivative,
        dtype=int,
    )

    # Compute coefficients for each derivative
    coefs = np.cumprod(coef_mat, axis=0)

    # Add derivative 0 row
    coefs = np.concatenate((np.ones((1, basis.n_basis)), coefs))

    # Now each row correspond to each basis and each column to
    # each derivative
    coefs_t = coefs.T

    # Multiply by the weights
    weighted_coefs = coefs_t * weights
    assert len(weighted_coefs) == basis.n_basis

    # Now each row has the right weight, but the polynomials are in a
    # decreasing order and with different exponents

    # Resize the coefs so that there are as many rows as the number of
    # basis
    # The matrix is now triangular
    # refcheck is False to prevent exceptions while debugging
    weighted_coefs = np.copy(weighted_coefs.T)
    weighted_coefs.resize(
        basis.n_basis,
        basis.n_basis,
        refcheck=False,
    )
    weighted_coefs = weighted_coefs.T

    # Shift the coefficients so that they correspond to the right
    # exponent
    indexes = np.tril_indices(basis.n_basis)
    polynomials = np.zeros_like(weighted_coefs)
    polynomials[
        indexes[0],
        indexes[1] - indexes[0] - 1,
    ] = weighted_coefs[indexes]

    # At this point, each row of the matrix correspond to a polynomial
    # that is the result of applying the linear differential operator
    # to each element of the basis

    return polynomials


@gramian_matrix_optimization.register
def monomial_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: Monomial,
) -> np.ndarray:
    """Optimized version for Monomial basis."""
    weights = linear_operator.constant_weights()
    if weights is None:
        return NotImplemented

    polynomials = _monomial_evaluate_constant_linear_diff_op(basis, weights)

    # Expand the polinomials with 0, so that the multiplication fits
    # inside. It will need the double of the degree
    length_with_padding = polynomials.shape[1] * 2 - 1

    # Multiplication of polynomials is a convolution.
    # The convolution can be performed in parallel applying a Fourier
    # transform and then doing a normal multiplication in that
    # space, coverting back with the inverse Fourier transform
    fft = np.fft.rfft(polynomials, length_with_padding)

    # We compute only the upper matrix, as the penalty matrix is
    # symmetrical
    indices = np.triu_indices(basis.n_basis)
    fft_mul = fft[indices[0]] * fft[indices[1]]

    integrand = np.fft.irfft(fft_mul, length_with_padding)

    integration_domain = basis.domain_range[0]

    # To integrate, divide by the position and increase the exponent
    # in the evaluation
    denom = np.arange(integrand.shape[1], 0, -1)
    integrand /= denom

    # Add column of zeros at the right to increase exponent
    integrand = np.pad(
        integrand,
        pad_width=(
            (0, 0),
            (0, 1),
        ),
        mode='constant',
    )

    # Now, apply Barrow's rule
    # polyval applies Horner method over the first dimension,
    # so we need to transpose
    x_right = np.polyval(integrand.T, integration_domain[1])
    x_left = np.polyval(integrand.T, integration_domain[0])

    integral = x_right - x_left

    penalty_matrix = np.empty((basis.n_basis, basis.n_basis))

    # Set upper matrix
    penalty_matrix[indices] = integral

    # Set lower matrix
    penalty_matrix[(indices[1], indices[0])] = integral

    return penalty_matrix


def _fourier_penalty_matrix_optimized_orthonormal(
    basis: Fourier,
    weights: np.ndarray,
) -> np.ndarray:
    """Return the penalty when the basis is orthonormal."""
    signs = np.array([1, 1, -1, -1])
    signs_expanded = np.tile(signs, len(weights) // 4 + 1)

    signs_odd = signs_expanded[:len(weights)]
    signs_even = signs_expanded[1:len(weights) + 1]

    phases = (
        np.arange(1, (basis.n_basis - 1) // 2 + 1)
        * 2 * np.pi / basis.period
    )

    # Compute increasing powers
    coefs_no_sign = np.vander(phases, len(weights), increasing=True)

    coefs_no_sign *= weights

    coefs_odd = signs_odd * coefs_no_sign
    coefs_even = signs_even * coefs_no_sign

    # After applying the linear differential operator to a sinusoidal
    # element of the basis e, the result can be expressed as
    # A e + B e*, where e* is the other basis element in the pair
    # with the same phase

    odd_sin_coefs = np.sum(coefs_odd[:, ::2], axis=1)
    odd_cos_coefs = np.sum(coefs_odd[:, 1::2], axis=1)

    even_cos_coefs = np.sum(coefs_even[:, ::2], axis=1)
    even_sin_coefs = np.sum(coefs_even[:, 1::2], axis=1)

    # The diagonal is the inner product of A e + B e*
    # with itself. As the basis is orthonormal, the cross products e e*
    # are 0, and the products e e and e* e* are one.
    # Thus, the diagonal is A^2 + B^2
    # All elements outside the main diagonal are 0
    main_diag_odd = odd_sin_coefs**2 + odd_cos_coefs**2
    main_diag_even = even_sin_coefs**2 + even_cos_coefs**2

    # The main diagonal should intercalate both diagonals
    main_diag = np.array((main_diag_odd, main_diag_even)).T.ravel()

    penalty_matrix = np.diag(main_diag)

    # Add row and column for the constant
    penalty_matrix = np.pad(
        penalty_matrix,
        pad_width=(
            (1, 0),
            (1, 0),
        ),
        mode='constant',
    )

    penalty_matrix[0, 0] = weights[0]**2

    return penalty_matrix


@gramian_matrix_optimization.register
def fourier_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: Fourier,
) -> np.ndarray:
    """Optimized version for Fourier basis."""
    weights = linear_operator.constant_weights()
    if weights is None:
        return NotImplemented

    # If the period and domain range are not the same, the basis functions
    # are not orthogonal
    if basis.period != (basis.domain_range[0][1] - basis.domain_range[0][0]):
        return NotImplemented

    return _fourier_penalty_matrix_optimized_orthonormal(basis, weights)


@gramian_matrix_optimization.register
def bspline_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: BSpline,
) -> np.ndarray:
    """Optimized version for BSpline basis."""
    coefs = linear_operator.constant_weights()
    if coefs is None:
        return NotImplemented

    nonzero = np.flatnonzero(coefs)

    # All derivatives above the order of the spline are effectively
    # zero
    nonzero = nonzero[nonzero < basis.order]

    if len(nonzero) == 0:
        return np.zeros((basis.n_basis, basis.n_basis))

    # We will only deal with one nonzero coefficient right now
    if len(nonzero) != 1:
        return NotImplemented

    derivative_degree = nonzero[0]

    if derivative_degree == basis.order - 1:
        # The derivative of the bsplines are constant in the intervals
        # defined between knots
        knots = np.array(basis.knots)
        mid_inter = (knots[1:] + knots[:-1]) / 2
        basis_deriv = basis.derivative(order=derivative_degree)
        constants = basis_deriv(mid_inter)[..., 0].T
        knots_intervals = np.diff(basis.knots)
        # Integration of product of constants
        return constants.T @ np.diag(knots_intervals) @ constants

    # We only deal with the case without zero length intervals
    # for now
    if np.any(np.diff(basis.knots) == 0):
        return NotImplemented

    # Compute exactly using the piecewise polynomial
    # representation of splines

    # Places m knots at the boundaries
    knots = np.array(basis._evaluation_knots())

    # c is used the select which spline the function
    # PPoly.from_spline below computes
    c = np.zeros(len(knots))

    # Initialise empty list to store the piecewise polynomials
    ppoly_lst = []

    no_0_intervals = np.where(np.diff(knots) > 0)[0]

    # For each basis gets its piecewise polynomial representation
    for i in range(basis.n_basis):

        # Write a 1 in c in the position of the spline
        # transformed in each iteration
        c[i] = 1

        # Gets the piecewise polynomial representation and gets
        # only the positions for no zero length intervals
        # This polynomial are defined relatively to the knots
        # meaning that the column i corresponds to the ith knot.
        # Let the ith knot be a
        # Then f(x) = pp(x - a)
        pp = PPoly.from_spline((knots, c, basis.order - 1))
        pp_coefs = pp.c[:, no_0_intervals]

        # We have the coefficients for each interval in coordinates
        # (x - a), so we will need to subtract a when computing the
        # definite integral
        ppoly_lst.append(pp_coefs)
        c[i] = 0

    # Now for each pair of basis computes the inner product after
    # applying the linear differential operator
    penalty_matrix = np.zeros((basis.n_basis, basis.n_basis))
    for interval, _ in enumerate(no_0_intervals):
        for i in range(basis.n_basis):
            poly_i = np.trim_zeros(ppoly_lst[i][:, interval], 'f')
            if len(poly_i) <= derivative_degree:
                # if the order of the polynomial is lesser or
                # equal to the derivative the result of the
                # integral will be 0
                continue
            # indefinite integral
            derivative = polyder(poly_i, derivative_degree)
            square = polymul(derivative, derivative)
            integral = polyint(square)

            # definite integral
            penalty_matrix[i, i] += np.diff(
                polyval(
                    integral,
                    basis.knots[interval: interval + 2]
                    - basis.knots[interval],
                ),
            )[0]

            for j in range(i + 1, basis.n_basis):
                poly_j = np.trim_zeros(ppoly_lst[j][:, interval], 'f')
                if len(poly_j) <= derivative_degree:
                    # if the order of the polynomial is lesser
                    # or equal to the derivative the result of
                    # the integral will be 0
                    continue
                    # indefinite integral
                integral = polyint(
                    polymul(
                        polyder(poly_i, derivative_degree),
                        polyder(poly_j, derivative_degree),
                    ),
                )
                # definite integral
                penalty_matrix[i, j] += np.diff(
                    polyval(
                        integral,
                        basis.knots[interval: interval + 2]
                        - basis.knots[interval],
                    ),
                )[0]
                penalty_matrix[j, i] = penalty_matrix[i, j]
    return penalty_matrix


@gramian_matrix_optimization.register
def fdatagrid_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: FDataGrid,
) -> np.ndarray:
    """Optimized version for FDatagrid."""
    evaluated_basis = sum(
        w(basis.grid_points[0]) if callable(w) else w
        * basis.derivative(order=i)(basis.grid_points[0])
        for i, w in enumerate(linear_operator.weights)
    )

    indices = np.triu_indices(basis.n_samples)
    product = evaluated_basis[indices[0]] * evaluated_basis[indices[1]]

    triang_vec = scipy.integrate.simps(product[..., 0], x=basis.grid_points)

    matrix = np.empty((basis.n_samples, basis.n_samples))

    # Set upper matrix
    matrix[indices] = triang_vec

    # Set lower matrix
    matrix[(indices[1], indices[0])] = triang_vec

    return matrix



from __future__ import annotations

from typing import Any, TypeVar

import numpy as np

from ...representation import FDataGrid
from ...representation._typing import Vector
from ...representation.basis import Basis
from ._operators import Operator, gramian_matrix_optimization

T = TypeVar("T", bound=Vector)


class Identity(Operator[T, T]):
    """Identity operator.

    Linear operator that returns its input.

    .. math::
        Ix = x

    Can be applied to both functional and multivariate data.

    """

    def __call__(self, f: T) -> T:  # noqa: D102
        return f


@gramian_matrix_optimization.register
def basis_penalty_matrix_optimized(
    linear_operator: Identity[Any],
    basis: Basis,
) -> np.ndarray:
    """Optimized version of the penalty matrix for Basis."""
    return basis.gram_matrix()


@gramian_matrix_optimization.register
def fdatagrid_penalty_matrix_optimized(
    linear_operator: Identity[Any],
    basis: FDataGrid,
) -> np.ndarray:
    """Optimized version of the penalty matrix for FDataGrid."""
    from ..metrics import l2_norm

    return np.diag(l2_norm(basis)**2)



# encoding: utf-8

"""
Functional Data Analysis Python package.

Functional Data Analysis, or FDA, is the field of Statistics that analyses
data that depend on a continuous parameter.

This package offers classes, methods and functions to give support to FDA
in Python. Includes a wide range of utils to work with functional data, and its
representation, exploratory analysis, or preprocessing, among other tasks
such as inference, classification, regression or clustering of functional data.
See documentation or visit the
`github page <https://github.com/GAA-UAM/scikit-fda>`_ of the project for
further information on the features included in the package.

The documentation is available at
`fda.readthedocs.io/en/stable/ <https://fda.readthedocs.io/en/stable/>`_, which
includes detailed information of the different modules, classes and methods of
the package, along with several examples showing different funcionalities.
"""

import os
import sys

from setuptools import find_packages, setup

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

DOCLINES = (__doc__ or '').split("\n")

with open(os.path.join(os.path.dirname(__file__),
                       'VERSION'), 'r') as version_file:
    version = version_file.read().strip()

setup(name='scikit-fda',
      version=version,
      description=DOCLINES[1],
      long_description="\n".join(DOCLINES[3:]),
      url='https://fda.readthedocs.io',
      maintainer='Carlos Ramos Carreño',
      maintainer_email='vnmabus@gmail.com',
      include_package_data=True,
      platforms=['any'],
      license='BSD',
      packages=find_packages(),
      python_requires='>=3.7, <4',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Typing :: Typed',
      ],
      install_requires=[
          'cython',
          'dcor',
          'fdasrsf>=2.2.0',
          'findiff',
          'matplotlib',
          'multimethod>=1.5',
          'numpy>=1.16',
          'pandas>=1.0',
          'rdata',
          'scikit-datasets[cran]>=0.1.24',
          'scikit-learn>=0.20',
          'scipy>=1.3.0',
          'typing-extensions',
      ],
      setup_requires=pytest_runner,
      tests_require=['pytest'],
      test_suite='tests',
      zip_safe=False)



"""Module with the extrapolation methods.

Defines methods to evaluate points outside the :term:`domain` range.

"""
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    NoReturn,
    Optional,
    Union,
    cast,
    overload,
)

import numpy as np
from typing_extensions import Literal

from ._typing import ArrayLike
from .evaluator import Evaluator

if TYPE_CHECKING:
    from . import FData

ExtrapolationLike = Union[
    Evaluator,
    Literal["bounds", "exception", "nan", "none", "periodic", "zeros"],
]


class PeriodicExtrapolation(Evaluator):
    """Extend the :term:`domain` range periodically.

    Examples:
        >>> from skfda.datasets import make_sinusoidal_process
        >>> from skfda.representation.extrapolation import (
        ...     PeriodicExtrapolation)
        >>> fd = make_sinusoidal_process(n_samples=2, random_state=0)

        We can set the default type of extrapolation

        >>> fd.extrapolation = PeriodicExtrapolation()
        >>> fd([-.5, 0, 1.5]).round(3)
        array([[[-0.724],
                [ 0.976],
                [-0.724]],
               [[-1.086],
                [ 0.759],
                [-1.086]]])

        This extrapolator is equivalent to the string `"periodic"`

        >>> fd.extrapolation = 'periodic'
        >>> fd([-.5, 0, 1.5]).round(3)
        array([[[-0.724],
                [ 0.976],
                [-0.724]],
               [[-1.086],
                [ 0.759],
                [-1.086]]])
    """

    def _evaluate(  # noqa: D102
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:

        domain_range = np.asarray(fdata.domain_range)

        # Extends the domain periodically in each dimension
        eval_points -= domain_range[:, 0]
        eval_points %= domain_range[:, 1] - domain_range[:, 0]
        eval_points += domain_range[:, 0]

        return fdata(eval_points, aligned=aligned)  # type: ignore


class BoundaryExtrapolation(Evaluator):
    """Extend the :term:`domain` range using the boundary values.

    Examples:
        >>> from skfda.datasets import make_sinusoidal_process
        >>> from skfda.representation.extrapolation import (
        ...     BoundaryExtrapolation)
        >>> fd = make_sinusoidal_process(n_samples=2, random_state=0)

        We can set the default type of extrapolation

        >>> fd.extrapolation = BoundaryExtrapolation()
        >>> fd([-.5, 0, 1.5]).round(3)
        array([[[ 0.976],
                [ 0.976],
                [ 0.797]],
               [[ 0.759],
                [ 0.759],
                [ 1.125]]])

        This extrapolator is equivalent to the string `"bounds"`.

        >>> fd.extrapolation = 'bounds'
        >>> fd([-.5, 0, 1.5]).round(3)
        array([[[ 0.976],
                [ 0.976],
                [ 0.797]],
               [[ 0.759],
                [ 0.759],
                [ 1.125]]])
    """

    def _evaluate(  # noqa: D102
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:

        domain_range = fdata.domain_range

        if aligned:
            eval_points = np.asarray(eval_points)

            for i in range(fdata.dim_domain):
                a, b = domain_range[i]
                eval_points[eval_points[..., i] < a, i] = a
                eval_points[eval_points[..., i] > b, i] = b
        else:
            eval_points = cast(Iterable[ArrayLike], eval_points)

            for points_per_sample in eval_points:

                points_per_sample = np.asarray(points_per_sample)

                for i in range(fdata.dim_domain):
                    a, b = domain_range[i]
                    points_per_sample[points_per_sample[..., i] < a, i] = a
                    points_per_sample[points_per_sample[..., i] > b, i] = b

        return fdata(eval_points, aligned=aligned)  # type: ignore


class ExceptionExtrapolation(Evaluator):
    """Raise an exception.

    Examples:
        >>> from skfda.datasets import make_sinusoidal_process
        >>> from skfda.representation.extrapolation import (
        ...     ExceptionExtrapolation)
        >>> fd = make_sinusoidal_process(n_samples=2, random_state=0)

        We can set the default type of extrapolation

        >>> fd.extrapolation = ExceptionExtrapolation()
        >>> try:
        ...     fd([-.5, 0, 1.5]).round(3)
        ... except ValueError as e:
        ...     print(e)
        Attempt to evaluate points outside the domain range.

        This extrapolator is equivalent to the string `"exception"`.

        >>> fd.extrapolation = 'exception'
        >>> try:
        ...     fd([-.5, 0, 1.5]).round(3)
        ... except ValueError as e:
        ...     print(e)
        Attempt to evaluate points outside the domain range.

    """

    def _evaluate(  # noqa: D102
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> NoReturn:

        raise ValueError(
            "Attempt to evaluate points outside the domain range.",
        )


class FillExtrapolation(Evaluator):
    """
    Values outside the :term:`domain` range will be filled with a fixed value.

    Examples:
        >>> from skfda.datasets import make_sinusoidal_process
        >>> from skfda.representation.extrapolation import FillExtrapolation
        >>> fd = make_sinusoidal_process(n_samples=2, random_state=0)

        We can set the default type of extrapolation

        >>> fd.extrapolation = FillExtrapolation(0)
        >>> fd([-.5, 0, 1.5]).round(3)
        array([[[ 0.   ],
                [ 0.976],
                [ 0.   ]],
               [[ 0.   ],
                [ 0.759],
                [ 0.   ]]])

        The previous extrapolator is equivalent to the string `"zeros"`.
        In the same way FillExtrapolation(np.nan) is equivalent to `"nan"`.

        >>> fd.extrapolation = "nan"
        >>> fd([-.5, 0, 1.5]).round(3)
        array([[[   nan],
                [ 0.976],
                [   nan]],
               [[   nan],
                [ 0.759],
                [   nan]]])
    """

    def __init__(self, fill_value: float) -> None:
        self.fill_value = fill_value

    def _fill(self, fdata: FData, eval_points: ArrayLike) -> np.ndarray:
        eval_points = np.asarray(eval_points)

        shape = (
            fdata.n_samples,
            eval_points.shape[-2],
            fdata.dim_codomain,
        )
        return np.full(shape, self.fill_value)

    def _evaluate(  # noqa: D102
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:
        from .._utils import _to_array_maybe_ragged

        if aligned:
            eval_points = cast(ArrayLike, eval_points)
            return self._fill(fdata, eval_points)

        eval_points = cast(Iterable[ArrayLike], eval_points)

        res_list = [self._fill(fdata, p) for p in eval_points]

        return _to_array_maybe_ragged(res_list)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"fill_value={self.fill_value})"
        )

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and (
                self.fill_value == other.fill_value
                # NaNs compare unequal. Should we distinguish between
                # different NaN types and payloads?
                or (np.isnan(self.fill_value) and np.isnan(other.fill_value))
            )
        )


@overload
def _parse_extrapolation(
    extrapolation: None,
) -> None:
    pass


@overload
def _parse_extrapolation(
    extrapolation: ExtrapolationLike,
) -> Evaluator:
    pass


def _parse_extrapolation(
    extrapolation: Optional[ExtrapolationLike],
) -> Optional[Evaluator]:
    """Parse the argument `extrapolation` of `FData`.

    If extrapolation is None returns the default extrapolator.

    Args:
        extrapolation (:class:´Extrapolator´, str or Callable): Argument
            extrapolation to be parsed.

    Returns:
        (:class:´Extrapolator´ or Callable): Extrapolation method.

    """
    if extrapolation is None:
        return None

    elif isinstance(extrapolation, str):
        return extrapolation_methods[extrapolation.lower()]

    return extrapolation


#: Dictionary with the extrapolation methods.
extrapolation_methods = {
    "bounds": BoundaryExtrapolation(),
    "exception": ExceptionExtrapolation(),
    "nan": FillExtrapolation(np.nan),
    "none": None,
    "periodic": PeriodicExtrapolation(),
    "zeros": FillExtrapolation(0),
}



from __future__ import annotations

import operator
from typing import (
    Any,
    Callable,
    Dict,
    NamedTuple,
    Optional,
    Tuple,
    Union,
    overload,
)

import numpy as np
import sklearn.base
import sklearn.utils.validation
from sklearn.feature_selection import (
    mutual_info_classif,
    mutual_info_regression,
)
from typing_extensions import Final, Literal

from ...._utils import RandomStateLike, _compute_dependence, _DependenceMeasure
from ....representation._typing import NDArrayFloat, NDArrayInt
from ....representation.grid import FDataGrid

_Criterion = Callable[[NDArrayFloat, NDArrayFloat], NDArrayFloat]
_CriterionLike = Union[
    _Criterion,
    Literal["difference", "quotient"],
]


class Method(NamedTuple):
    """Predefined mRMR method."""

    relevance_dependence_measure: _DependenceMeasure
    redundancy_dependence_measure: _DependenceMeasure
    criterion: _Criterion


def mutual_information(
    x: NDArrayFloat,
    y: Union[NDArrayInt, NDArrayFloat],
    n_neighbors: Optional[int] = None,
    random_state: RandomStateLike = None,
) -> NDArrayFloat:
    """Compute mutual information."""
    y = y.ravel()

    method = (
        mutual_info_regression if issubclass(y.dtype.type, np.floating)
        else mutual_info_classif
    )

    extra_args: Dict[str, Any] = {}
    if n_neighbors is not None:
        extra_args['n_neighbors'] = n_neighbors

    return method(x, y, random_state=random_state, **extra_args)


MID: Final = Method(
    relevance_dependence_measure=mutual_information,
    redundancy_dependence_measure=mutual_information,
    criterion=operator.sub,
)


MIQ: Final = Method(
    relevance_dependence_measure=mutual_information,
    redundancy_dependence_measure=mutual_information,
    criterion=operator.truediv,
)


MethodName = Literal["MID", "MIQ"]


def _parse_method(name: MethodName) -> Method:
    if name == "MID":
        return MID
    elif name == "MIQ":
        return MIQ


def _mrmr(
    X: NDArrayFloat,
    Y: Union[NDArrayInt, NDArrayFloat],
    n_features_to_select: int = 1,
    relevance_dependence_measure: _DependenceMeasure = mutual_information,
    redundancy_dependence_measure: _DependenceMeasure = mutual_information,
    criterion: _Criterion = operator.truediv,
) -> Tuple[NDArrayInt, NDArrayFloat, NDArrayFloat]:
    indexes = list(range(X.shape[1]))

    selected_features = []
    scores = []
    selected_relevances = []

    relevances = _compute_dependence(
        X[..., np.newaxis],
        Y,
        dependence_measure=relevance_dependence_measure,
    )
    redundancies = np.zeros((X.shape[1], X.shape[1]))

    max_index = int(np.argmax(relevances))
    selected_features.append(indexes[max_index])
    scores.append(relevances[max_index])
    selected_relevances.append(relevances[max_index])

    indexes.remove(max_index)

    # TODO: Vectorize
    for i in range(1, n_features_to_select):

        # Calculate redundancies of the last selected variable
        last_selected = selected_features[i - 1]

        for j in range(X.shape[1]):
            if not redundancies[last_selected, j]:
                redundancies[last_selected, j] = redundancy_dependence_measure(
                    X[:, last_selected, np.newaxis],
                    X[:, j, np.newaxis],
                )
                redundancies[j, last_selected] = redundancies[last_selected, j]

        W = np.mean(
            redundancies[np.ix_(selected_features[:i], indexes)],
            axis=0,
        )

        coef = criterion(relevances[indexes], W)

        max_index = int(np.argmax(coef))
        selected_features.append(indexes[max_index])
        scores.append(coef[max_index])
        selected_relevances.append(relevances[max_index])

        indexes.remove(indexes[max_index])

    return (
        np.asarray(selected_features),
        np.asarray(scores),
        np.asarray(relevances),
    )


class MinimumRedundancyMaximumRelevance(
    sklearn.base.BaseEstimator,  # type: ignore
    sklearn.base.TransformerMixin,  # type: ignore
):
    r"""
    Minimum redundancy maximum relevance (mRMR) method.

    This is a greedy version of mRMR that selects the variables iteratively.
    This method considers the relevance of a variable as well as its redundancy
    with respect of the already selected ones.    

    It uses a dependence measure between random variables to compute the
    dependence between the candidate variable and the target (for the
    relevance) and another to compute the dependence between two variables
    (for the redundancy).
    It combines both measurements using a criterion such as the difference or
    the quotient, and then selects the variable that maximizes that quantity.
    For example, using the quotient criterion and the same dependence function
    :math:`D` for relevance and redundancy, the variable selected at the
    :math:`i`-th step would be :math:`X(t_i)` with

    .. math::
        t_i = \underset {t}{\operatorname {arg\,max}} \frac{D(X(t), y)}
        {\frac{1}{i-1}\sum_{j < i} D(X(t), X(t_j))}.

    For further discussion of the applicability of this method to functional
    data see :footcite:`berrendero++_2016_mrmr`.

    Parameters:
        n_features_to_select: Number of features to select.
        method: Predefined method to use (MID or MIQ).
        dependence_measure: Dependence measure to use both for relevance and
            for redundancy.
        relevance_dependence_measure: Dependence measure used to compute
            relevance.
        redundancy_dependence_measure: Dependence measure used to compute
            redundancy.
        criterion: Criterion to combine relevance and redundancy. It must be
            a Python callable with two inputs. As common choices include the
            difference and the quotient, both can be especified as strings.

    Examples:
        >>> from skfda.preprocessing.dim_reduction import variable_selection
        >>> from skfda.datasets import make_gaussian_process
        >>> import skfda
        >>> import numpy as np
        >>> import operator
        >>> import dcor

        We create trajectories from two classes, one with zero mean and the
        other with a peak-like mean. Both have Brownian covariance.

        >>> n_samples = 1000
        >>> n_features = 100
        >>>
        >>> def mean_1(t):
        ...     return (
        ...         np.abs(t - 0.25)
        ...         - 2 * np.abs(t - 0.5)
        ...         + np.abs(t - 0.75)
        ...     )
        >>>
        >>> X_0 = make_gaussian_process(
        ...     n_samples=n_samples // 2,
        ...     n_features=n_features,
        ...     random_state=0,
        ... )
        >>> X_1 = make_gaussian_process(
        ...     n_samples=n_samples // 2,
        ...     n_features=n_features,
        ...     mean=mean_1,
        ...     random_state=1,
        ... )
        >>> X = skfda.concatenate((X_0, X_1))
        >>>
        >>> y = np.zeros(n_samples, dtype=np.int_)
        >>> y [n_samples // 2:] = 1

        Select the relevant points to distinguish the two classes. You
        may specify a method such as MIQ (the default) or MID.

        >>> mrmr = variable_selection.MinimumRedundancyMaximumRelevance(
        ...     n_features_to_select=3,
        ...     method="MID",
        ... )
        >>> _ = mrmr.fit(X, y)
        >>> point_mask = mrmr.get_support()
        >>> points = X.grid_points[0][point_mask]

        Apply the learned dimensionality reduction

        >>> X_dimred = mrmr.transform(X)
        >>> len(X.grid_points[0])
        100
        >>> X_dimred.shape
        (1000, 3)

        It is also possible to specify the measure of dependence used (or
        even different ones for relevance and redundancy) as well as the
        function to combine relevance and redundancy (usually the division
        or subtraction operations).

        >>> mrmr = variable_selection.MinimumRedundancyMaximumRelevance(
        ...     n_features_to_select=3,
        ...     dependence_measure=dcor.u_distance_correlation_sqr,
        ...     criterion="quotient",
        ... )
        >>> _ = mrmr.fit(X, y)

        As a toy example illustrating the customizability of this method,
        consider the following:

        >>> mrmr = variable_selection.MinimumRedundancyMaximumRelevance(
        ...     n_features_to_select=3,
        ...     relevance_dependence_measure=dcor.u_distance_covariance_sqr,
        ...     redundancy_dependence_measure=dcor.u_distance_correlation_sqr,
        ...     criterion=lambda rel, red: 0.5 * rel / red,
        ... )
        >>> _ = mrmr.fit(X, y)

    References:
        .. footbibliography::

    """

    @overload
    def __init__(
        self,
        *,
        n_features_to_select: int = 1,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        n_features_to_select: int = 1,
        method: Union[Method, MethodName],
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        n_features_to_select: int = 1,
        dependence_measure: _DependenceMeasure,
        criterion: _CriterionLike,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        n_features_to_select: int = 1,
        relevance_dependence_measure: _DependenceMeasure,
        redundancy_dependence_measure: _DependenceMeasure,
        criterion: _CriterionLike,
    ) -> None:
        pass

    def __init__(
        self,
        *,
        n_features_to_select: int = 1,
        method: Union[Method, MethodName, None] = None,
        dependence_measure: Optional[_DependenceMeasure] = None,
        relevance_dependence_measure: Optional[_DependenceMeasure] = None,
        redundancy_dependence_measure: Optional[_DependenceMeasure] = None,
        criterion: Optional[_CriterionLike] = None,
    ) -> None:
        self.n_features_to_select = n_features_to_select
        self.method = method
        self.dependence_measure = dependence_measure
        self.relevance_dependence_measure = relevance_dependence_measure
        self.redundancy_dependence_measure = redundancy_dependence_measure
        self.criterion = criterion

    def _validate_parameters(self) -> None:
        method = MIQ if all(
            p is None for p in (
                self.method,
                self.dependence_measure,
                self.relevance_dependence_measure,
                self.redundancy_dependence_measure,
                self.criterion,
            )
        ) else self.method

        if method:
            if (
                self.dependence_measure
                or self.relevance_dependence_measure
                or self.redundancy_dependence_measure
                or self.criterion
            ):
                raise ValueError(
                    "The 'method' parameter and the parameters "
                    "'dependency_measure', 'relevance_dependence_measure' "
                    "'redundancy_dependence_measure' and 'criterion' are "
                    "incompatible",
                )

            method = (
                _parse_method(method)
                if isinstance(method, str) else method
            )

            self.relevance_dependence_measure_ = (
                method.relevance_dependence_measure
            )
            self.redundancy_dependence_measure_ = (
                method.redundancy_dependence_measure
            )
            self.criterion_ = method.criterion

        else:
            if self.criterion is None:
                raise ValueError(
                    "You must specify a criterion parameter",
                )

            if self.criterion == "difference":
                self.criterion = operator.sub
            elif self.criterion == "quotient":
                self.criterion_ = operator.truediv
            else:
                self.criterion_ = self.criterion

            if self.dependence_measure:
                if (
                    self.relevance_dependence_measure
                    or self.redundancy_dependence_measure
                ):
                    raise ValueError(
                        "The 'dependency_measure' parameter and the "
                        "parameters 'relevance_dependence_measure' "
                        "and 'redundancy_dependence_measure' "
                        "are incompatible",
                    )

                self.relevance_dependence_measure_ = (
                    self.dependence_measure
                )
                self.redundancy_dependence_measure_ = (
                    self.dependence_measure
                )
            else:
                if not self.relevance_dependence_measure:
                    raise ValueError(
                        "Missing parameter 'relevance_dependence_measure'",
                    )
                if not self.redundancy_dependence_measure:
                    raise ValueError(
                        "Missing parameter 'redundancy_dependence_measure'",
                    )
                self.relevance_dependence_measure_ = (
                    self.relevance_dependence_measure
                )
                self.redundancy_dependence_measure_ = (
                    self.redundancy_dependence_measure
                )

    def fit(
        self,
        X: FDataGrid,
        y: Union[NDArrayInt, NDArrayFloat],
    ) -> MinimumRedundancyMaximumRelevance:

        self._validate_parameters()

        X_array = X.data_matrix[..., 0]

        X_array, y = sklearn.utils.validation.check_X_y(X_array, y)

        self.features_shape_ = X_array.shape[1:]

        self.results_ = _mrmr(
            X=X_array,
            Y=y,
            n_features_to_select=self.n_features_to_select,
            relevance_dependence_measure=self.relevance_dependence_measure_,
            redundancy_dependence_measure=self.redundancy_dependence_measure_,
            criterion=self.criterion_,
        )[0]

        return self

    def transform(
        self,
        X: FDataGrid,
        y: None = None,
    ) -> NDArrayFloat:

        X_array = X.data_matrix[..., 0]

        sklearn.utils.validation.check_is_fitted(self)

        X_array = sklearn.utils.validation.check_array(X_array)

        if X_array.shape[1:] != self.features_shape_:
            raise ValueError(
                "The trajectories have a different number of "
                "points than the ones fitted",
            )

        return X_array[:, self.results_]

    def get_support(self, indices: bool = False) -> NDArrayInt:
        indexes_unraveled = self.results_
        if indices:
            return indexes_unraveled

        mask = np.zeros(self.features_shape_[0], dtype=bool)
        mask[self.results_] = True
        return mask



from typing import Any, Optional, Sequence, Tuple, Type, TypeVar

import numpy as np
import scipy.interpolate
from numpy import polyint, polymul, polyval
from scipy.interpolate import BSpline as SciBSpline, PPoly

from ..._utils import _to_domain_range
from .._typing import DomainRangeLike
from ._basis import Basis

T = TypeVar("T", bound='BSpline')


class BSpline(Basis):
    r"""BSpline basis.

    BSpline basis elements are defined recursively as:

    .. math::
        B_{i, 1}(x) = 1 \quad \text{if } t_i \le x < t_{i+1},
        \quad 0 \text{ otherwise}

    .. math::
        B_{i, k}(x) = \frac{x - t_i}{t_{i+k} - t_i} B_{i, k-1}(x)
        + \frac{t_{i+k+1} - x}{t_{i+k+1} - t_{i+1}} B_{i+1, k-1}(x)

    Where k indicates the order of the spline.

    Implementation details: In order to allow a discontinuous behaviour at
    the boundaries of the domain it is necessary to placing m knots at the
    boundaries [RS05]_. This is automatically done so that the user only has to
    specify a single knot at the boundaries.

    Parameters:
        domain_range: A tuple of length 2 containing the initial and
            end values of the interval over which the basis can be evaluated.
        n_basis: Number of functions in the basis.
        order: Order of the splines. One greather than their degree.
        knots: List of knots of the spline functions.

    Examples:
        Constructs specifying number of basis and order.

        >>> bss = BSpline(n_basis=8, order=4)

        If no order is specified defaults to 4 because cubic splines are
        the most used. So the previous example is the same as:

        >>> bss = BSpline(n_basis=8)

        It is also possible to create a BSpline basis specifying the knots.

        >>> bss = BSpline(knots=[0, 0.2, 0.4, 0.6, 0.8, 1])

        Once we create a basis we can evaluate each of its functions at a
        set of points.

        >>> bss = BSpline(n_basis=3, order=3)
        >>> bss([0, 0.5, 1])
        array([[[ 1.  ],
                [ 0.25],
                [ 0.  ]],
               [[ 0.  ],
                [ 0.5 ],
                [ 0.  ]],
               [[ 0.  ],
                [ 0.25],
                [ 1.  ]]])

        And evaluates first derivative

        >>> deriv = bss.derivative()
        >>> deriv([0, 0.5, 1])
        array([[[-2.],
                [-1.],
                [ 0.]],
               [[ 2.],
                [ 0.],
                [-2.]],
               [[ 0.],
                [ 1.],
                [ 2.]]])

    References:
        .. [RS05] Ramsay, J., Silverman, B. W. (2005). *Functional Data
            Analysis*. Springer. 50-51.

    """

    def __init__(
        self,
        domain_range: Optional[DomainRangeLike] = None,
        n_basis: Optional[int] = None,
        order: int = 4,
        knots: Optional[Sequence[float]] = None,
    ) -> None:
        """Bspline basis constructor."""
        if domain_range is not None:
            domain_range = _to_domain_range(domain_range)

            if len(domain_range) != 1:
                raise ValueError("Domain range should be unidimensional.")

            domain_range = domain_range[0]

        # Knots default to equally space points in the domain_range
        if knots is not None:
            knots = tuple(knots)
            knots = sorted(knots)
            if domain_range is None:
                domain_range = (knots[0], knots[-1])
            elif domain_range[0] != knots[0] or domain_range[1] != knots[-1]:
                raise ValueError(
                    "The ends of the knots must be the same "
                    "as the domain_range.",
                )

        # n_basis default to number of knots + order of the splines - 2
        if n_basis is None:
            if knots is None:
                raise ValueError(
                    "Must provide either a list of knots or the"
                    "number of basis.",
                )
            n_basis = len(knots) + order - 2

        if (n_basis - order + 2) < 2:
            raise ValueError(
                f"The number of basis ({n_basis}) minus the "
                f"order of the bspline ({order}) should be "
                f"greater than 3.",
            )

        self._order = order
        self._knots = None if knots is None else tuple(knots)
        super().__init__(domain_range=domain_range, n_basis=n_basis)

        # Checks
        if self.n_basis != self.order + len(self.knots) - 2:
            raise ValueError(
                f"The number of basis ({self.n_basis}) has to "
                f"equal the order ({self.order}) plus the "
                f"number of knots ({len(self.knots)}) minus 2.",
            )

    @property
    def knots(self) -> Tuple[float, ...]:
        if self._knots is None:
            return tuple(np.linspace(
                *self.domain_range[0],
                self.n_basis - self.order + 2,
            ))

        return self._knots

    @property
    def order(self) -> int:
        return self._order

    def _evaluation_knots(self) -> Tuple[float, ...]:
        """
        Get the knots adding m knots to the boundary.

        This needs to be done in order to allow a discontinuous behaviour
        at the boundaries of the domain [RS05]_.

        References:
            .. [RS05] Ramsay, J., Silverman, B. W. (2005). *Functional Data
                Analysis*. Springer. 50-51.
        """
        return tuple(
            (self.knots[0],) * (self.order - 1) + self.knots
            + (self.knots[-1],) * (self.order - 1),
        )

    def _evaluate(self, eval_points: np.ndarray) -> np.ndarray:

        # Input is scalar
        eval_points = eval_points[..., 0]

        # Places m knots at the boundaries
        knots = self._evaluation_knots()

        # c is used the select which spline the function splev below computes
        c = np.zeros(len(knots))

        # Initialise empty matrix
        mat = np.empty((self.n_basis, len(eval_points)))

        # For each basis computes its value for each evaluation point
        for i in range(self.n_basis):
            # write a 1 in c in the position of the spline calculated in each
            # iteration
            c[i] = 1
            # compute the spline
            mat[i] = scipy.interpolate.splev(
                eval_points,
                (knots, c, self.order - 1),
            )
            c[i] = 0

        return mat

    def _derivative_basis_and_coefs(
        self: T,
        coefs: np.ndarray,
        order: int = 1,
    ) -> Tuple[T, np.ndarray]:
        if order >= self.order:
            return (
                type(self)(n_basis=1, domain_range=self.domain_range, order=1),
                np.zeros((len(coefs), 1)),
            )

        deriv_splines = [
            self._to_scipy_bspline(coefs[i]).derivative(order)
            for i in range(coefs.shape[0])
        ]

        deriv_coefs = [
            self._from_scipy_bspline(spline)[1]
            for spline in deriv_splines
        ]

        deriv_basis = self._from_scipy_bspline(deriv_splines[0])[0]

        return deriv_basis, np.array(deriv_coefs)[:, 0:deriv_basis.n_basis]

    def rescale(  # noqa: D102
        self: T,
        domain_range: Optional[DomainRangeLike] = None,
    ) -> T:

        knots = np.array(self.knots, dtype=np.dtype('float'))

        if domain_range is not None:  # Rescales the knots
            domain_range = _to_domain_range(domain_range)[0]
            knots -= knots[0]
            knots *= (
                (domain_range[1] - domain_range[0])
                / (self.knots[-1] - self.knots[0])
            )
            knots += domain_range[0]

            # Fix possible round error
            knots[0] = domain_range[0]
            knots[-1] = domain_range[1]

        else:
            # TODO: Allow multiple dimensions
            domain_range = self.domain_range[0]

        return type(self)(domain_range, self.n_basis, self.order, knots)

    def __repr__(self) -> str:
        """Representation of a BSpline basis."""
        return (
            f"{self.__class__.__name__}(domain_range={self.domain_range}, "
            f"n_basis={self.n_basis}, order={self.order}, "
            f"knots={self.knots})"
        )

    def _gram_matrix(self) -> np.ndarray:
        # Places m knots at the boundaries
        knots = self._evaluation_knots()

        # c is used the select which spline the function
        # PPoly.from_spline below computes
        c = np.zeros(len(knots))

        # Initialise empty list to store the piecewise polynomials
        ppoly_lst = []

        no_0_intervals = np.where(np.diff(knots) > 0)[0]

        # For each basis gets its piecewise polynomial representation
        for n in range(self.n_basis):

            # Write a 1 in c in the position of the spline
            # transformed in each iteration
            c[n] = 1

            # Gets the piecewise polynomial representation and gets
            # only the positions for no zero length intervals
            # This polynomial are defined relatively to the knots
            # meaning that the column i corresponds to the ith knot.
            # Let the ith knot be a
            # Then f(x) = pp(x - a)
            pp = PPoly.from_spline((knots, c, self.order - 1))
            pp_coefs = pp.c[:, no_0_intervals]

            # We have the coefficients for each interval in coordinates
            # (x - a), so we will need to subtract a when computing the
            # definite integral
            ppoly_lst.append(pp_coefs)
            c[n] = 0

        # Now for each pair of basis computes the inner product after
        # applying the linear differential operator
        matrix = np.zeros((self.n_basis, self.n_basis))

        for interval, _ in enumerate(no_0_intervals):
            for i in range(self.n_basis):
                poly_i = np.trim_zeros(
                    ppoly_lst[i][:, interval],
                    'f',
                )
                # Indefinite integral
                square = polymul(poly_i, poly_i)
                integral = polyint(square)

                # Definite integral
                matrix[i, i] += np.diff(
                    polyval(
                        integral, np.array(self.knots[interval: interval + 2])
                        - self.knots[interval],
                    ),
                )[0]

                # The Gram matrix is banded, so not all intervals are used
                for j in range(i + 1, min(i + self.order, self.n_basis)):
                    poly_j = np.trim_zeros(ppoly_lst[j][:, interval], 'f')

                    # Indefinite integral
                    integral = polyint(polymul(poly_i, poly_j))

                    # Definite integral
                    matrix[i, j] += np.diff(
                        polyval(
                            integral,
                            np.array(self.knots[interval: interval + 2])
                            - self.knots[interval],
                        ),
                    )[0]

                    # The matrix is symmetric
                    matrix[j, i] = matrix[i, j]

        return matrix

    def _to_scipy_bspline(self, coefs: np.ndarray) -> SciBSpline:

        knots = np.concatenate((
            np.repeat(self.knots[0], self.order - 1),
            self.knots,
            np.repeat(self.knots[-1], self.order - 1),
        ))

        return SciBSpline(knots, coefs, self.order - 1)

    @classmethod
    def _from_scipy_bspline(
        cls: Type[T],
        bspline: SciBSpline,
    ) -> Tuple[T, np.ndarray]:
        order = bspline.k
        knots = bspline.t

        # Remove additional knots at the borders
        if order != 0:
            knots = knots[order: -order]

        coefs = bspline.c
        domain_range = [knots[0], knots[-1]]

        return cls(domain_range, order=order + 1, knots=knots), coefs

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and self.order == other.order
            and self.knots == other.knots
        )

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.order, self.knots))



"""Neighbor models for supervised classification."""

from sklearn.base import ClassifierMixin
from sklearn.neighbors import (
    KNeighborsClassifier as _KNeighborsClassifier,
    RadiusNeighborsClassifier as _RadiusNeighborsClassifier,
)

from .._neighbors_base import (
    KNeighborsMixin,
    NeighborsBase,
    NeighborsClassifierMixin,
    NeighborsMixin,
    RadiusNeighborsMixin,
)


class KNeighborsClassifier(
    NeighborsBase,
    NeighborsMixin,
    KNeighborsMixin,
    ClassifierMixin,
    NeighborsClassifierMixin,
):
    """Classifier implementing the k-nearest neighbors vote.

    Parameters:
        n_neighbors (int, default = 5):
            Number of neighbors to use by default for :meth:`kneighbors`
            queries.
        weights (str or callable, default = 'uniform'):
            Weight function used in prediction.
            Possible values:
            - 'uniform': uniform weights. All points in each neighborhood
            are weighted equally.
            - 'distance': weight points by the inverse of their distance.
            in this case, closer neighbors of a query point will have a
            greater influence than neighbors which are further away.
            - [callable]: a user-defined function which accepts an
            array of distances, and returns an array of the same shape
            containing the weights.
        algorithm (string, optional):
            Algorithm used to compute the nearest neighbors:
            - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate algorithm
            based on the values passed to :meth:`fit` method.
        leaf_size (int, default = 30):
            Leaf size passed to BallTree or KDTree. This can affect the
            speed of the construction and query, as well as the memory
            required to store the tree. The optimal value depends on the
            nature of the problem.
        metric (string or callable, default
            :func:`l2_distance <skfda.misc.metrics.l2_distance>`):
            the distance metric to use for the tree. The default metric is
            the L2 distance. See the documentation of the metrics module
            for a list of available metrics.
        metric_params (dict, optional):
            Additional keyword arguments for the metric function.
        n_jobs (int or None, optional):
            The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context.
            ``-1`` means using all processors.
            Doesn't affect :meth:`fit` method.
        multivariate_metric (boolean, default = False):
            Indicates if the metric used is a sklearn distance between vectors
            (see :class:`~sklearn.neighbors.DistanceMetric`) or a functional
            metric of the module `skfda.misc.metrics` if ``False``.

    Examples:
        Firstly, we will create a toy dataset with 2 classes

        >>> from skfda.datasets import make_sinusoidal_process
        >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
        >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
        ...                               phase_std=.25, random_state=0)
        >>> fd = fd1.concatenate(fd2)
        >>> y = 15*[0] + 15*[1]

        We will fit a K-Nearest Neighbors classifier

        >>> from skfda.ml.classification import KNeighborsClassifier
        >>> neigh = KNeighborsClassifier()
        >>> neigh.fit(fd, y)
        KNeighborsClassifier(...)

        We can predict the class of new samples

        >>> neigh.predict(fd[::2]) # Predict labels for even samples
        array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1])

        And the estimated probabilities.

        >>> neigh.predict_proba(fd[0]) # Probabilities of sample 0
        array([[ 1.,  0.]])

    See also:
        :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
        :class:`~skfda.ml.classification.NearestCentroid`
        :class:`~skfda.ml.regression.KNeighborsRegressor`
        :class:`~skfda.ml.regression.RadiusNeighborsRegressor`
        :class:`~skfda.ml.clustering.NearestNeighbors`

    Notes:
        See Nearest Neighbors in the sklearn online documentation for a
        discussion of the choice of ``algorithm`` and ``leaf_size``.

        This class wraps the sklearn classifier
        `sklearn.neighbors.KNeighborsClassifier`.

    Warning:
        Regarding the Nearest Neighbors algorithms, if it is found that two
        neighbors, neighbor `k+1` and `k`, have identical distances
        but different labels, the results will depend on the ordering of the
        training data.

        https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm
    """

    def __init__(
        self,
        n_neighbors=5,
        weights='uniform',
        algorithm='auto',
        leaf_size=30,
        metric='l2',
        metric_params=None,
        n_jobs=1,
        multivariate_metric=False,
    ):
        super().__init__(
            n_neighbors=n_neighbors,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            metric_params=metric_params,
            n_jobs=n_jobs,
            multivariate_metric=multivariate_metric,
        )

    def predict_proba(self, X):
        """Calculate probability estimates for the test data X.

        Args:
            X (:class:`FDataGrid` or array-like): FDataGrid with the test
                samples or array (n_query, n_indexed) if metric ==
                'precomputed'.

        Returns:
            p (array of shape = (n_samples, n_classes), or a list of n_outputs
                of such arrays if n_outputs > 1):
                The class probabilities of the input samples. Classes are
                ordered by lexicographic order.
        """
        self._check_is_fitted()

        X = self._transform_to_multivariate(X)

        return self.estimator_.predict_proba(X)

    def _init_estimator(self, sklearn_metric):
        """Initialize the sklearn K neighbors estimator.

        Args:
            sklearn_metric (pyfunc or 'precomputed'): Metric compatible with
                sklearn API or matrix (n_samples, n_samples) with precomputed
                distances.

        Returns:
            Sklearn K Neighbors estimator initialized.
        """
        return _KNeighborsClassifier(
            n_neighbors=self.n_neighbors,
            weights=self.weights,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric=sklearn_metric,
            metric_params=self.metric_params,
            n_jobs=self.n_jobs,
        )


class RadiusNeighborsClassifier(
    NeighborsBase,
    NeighborsMixin,
    RadiusNeighborsMixin,
    ClassifierMixin,
    NeighborsClassifierMixin,
):
    """Classifier implementing a vote among neighbors within a given radius.

    Parameters:
        radius (float, default = 1.0):
            Range of parameter space to use by default for
            :meth:`radius_neighbors` queries.
        weights (str or callable, default = 'uniform'):
            Weight function used in prediction.
            Possible values:
            - 'uniform': uniform weights. All points in each neighborhood
            are weighted equally.
            - 'distance': weight points by the inverse of their distance.
            in this case, closer neighbors of a query point will have a
            greater influence than neighbors which are further away.
            - [callable]: a user-defined function which accepts an
            array of distances, and returns an array of the same shape
            containing the weights.
        algorithm (string, optional):
            Algorithm used to compute the nearest neighbors:
            - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate algorithm
            based on the values passed to :meth:`fit` method.
        leaf_size (int, default = 30):
            Leaf size passed to BallTree or KDTree. This can affect the
            speed of the construction and query, as well as the memory
            required to store the tree. The optimal value depends on the
            nature of the problem.
        metric (string or callable, default
            :func:`l2_distance <skfda.misc.metrics.l2_distance>`):
            the distance metric to use for the tree. The default metric is
            the L2 distance. See the documentation of the metrics module
            for a list of available metrics.
        outlier_label (int, optional):
            Label, which is given for outlier samples (samples with no
            neighbors on given radius).
            If set to None, ValueError is raised, when outlier is detected.
        metric_params (dict, optional):
            Additional keyword arguments for the metric function.
        n_jobs (int or None, optional):
            The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context.
            ``-1`` means using all processors.
        multivariate_metric (boolean, default = False):
            Indicates if the metric used is a sklearn distance between vectors
            (see :class:`~sklearn.neighbors.DistanceMetric`) or a functional
            metric of the module `skfda.misc.metrics` if ``False``.

    Examples:
        Firstly, we will create a toy dataset with 2 classes.

        >>> from skfda.datasets import make_sinusoidal_process
        >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
        >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
        ...                               phase_std=.25, random_state=0)
        >>> fd = fd1.concatenate(fd2)
        >>> y = 15*[0] + 15*[1]

        We will fit a Radius Nearest Neighbors classifier.

        >>> from skfda.ml.classification import RadiusNeighborsClassifier
        >>> neigh = RadiusNeighborsClassifier(radius=.3)
        >>> neigh.fit(fd, y)
        RadiusNeighborsClassifier(...radius=0.3...)

        We can predict the class of new samples.

        >>> neigh.predict(fd[::2]) # Predict labels for even samples
        array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1])

    See also:
        :class:`~skfda.ml.classification.KNeighborsClassifier`
        :class:`~skfda.ml.classification.NearestCentroid`
        :class:`~skfda.ml.regression.KNeighborsRegressor`
        :class:`~skfda.ml.regression.RadiusNeighborsRegressor`
        :class:`~skfda.ml.clustering.NearestNeighbors`

    Notes:
        See Nearest Neighbors in the sklearn online documentation for a
        discussion of the choice of ``algorithm`` and ``leaf_size``.

        This class wraps the sklearn classifier
        `sklearn.neighbors.RadiusNeighborsClassifier`.

        https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm
    """

    def __init__(
        self,
        radius=1.0,
        weights='uniform',
        algorithm='auto',
        leaf_size=30,
        metric='l2',
        metric_params=None,
        outlier_label=None,
        n_jobs=1,
        multivariate_metric=False,
    ):
        super().__init__(
            radius=radius,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            metric_params=metric_params,
            n_jobs=n_jobs,
            multivariate_metric=multivariate_metric,
        )

        self.outlier_label = outlier_label

    def _init_estimator(self, sklearn_metric):
        """Initialize the sklearn radius neighbors estimator.

        Args:
            sklearn_metric (pyfunc or 'precomputed'): Metric compatible with
                sklearn API or matrix (n_samples, n_samples) with precomputed
                distances.

        Returns:
            Sklearn Radius Neighbors estimator initialized.
        """
        return _RadiusNeighborsClassifier(
            radius=self.radius,
            weights=self.weights,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric=sklearn_metric,
            metric_params=self.metric_params,
            outlier_label=self.outlier_label,
            n_jobs=self.n_jobs,
        )
